import json
import time
from copy import deepcopy
from pathlib import Path

import numpy as np
import torch
from hydra.utils import instantiate
from nnmnkwii.io import hts
from nnmnkwii.preprocessing.f0 import interp1d
from nnsvs.gen import (
    postprocess_acoustic,
    postprocess_duration,
    postprocess_waveform,
    predict_acoustic,
    predict_duration,
    predict_timelag,
    predict_waveform,
)
from nnsvs.io.hts import (
    full_to_mono,
    get_note_indices,
    get_pitch_index,
    get_pitch_indices,
    label2phrases,
    label2phrases_str,
    segment_labels,
)
from nnsvs.logger import getLogger
from nnsvs.usfgan import USFGANWrapper
from nnsvs.util import MinMaxScaler, StandardScaler, extract_static_scaler, load_vocoder
from omegaconf import OmegaConf


class BaseSVS(object):
    """Base class for singing voice synthesis (SVS) inference

    All SVS engines should inherit from this class.

    The input of the SVS engine uses the HTS-style full-context labels.
    The output should be a tuple of raw waveform and sampling rate.
    To allow language-independent SVS, this base class does not define
    the interface for the frontend functionality such as
    converting musicXML/UST to HTS labels. The frontend processing
    should be done externally (e.g., using pysinsy or utaupy) or can
    be implemented with an optional method.
    """

    def svs(self, labels, *args, **kwargs):
        """Run SVS inference and returns the synthesized waveform

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels

        Returns:
            tuple: (waveform, sampling rate)
        """
        pass


class SPSVS(BaseSVS):
    """Statistical parametric singing voice synthesis (SPSVS)

    Use the ``svs`` method for the simplest inference, or use the
    separated methods (e.g.,``predict_acoustic`` and ``predict_waveform``)
    to control each components of the parametric SVS system.

    Args:
        model_dir (str): directory of the model
        device (str): cpu or cuda
        verbose (int): verbosity level

    Examples:

    Synthesize wavefrom from a musicxml file

    .. plot::

        import numpy as np
        import pysinsy
        from nnmnkwii.io import hts
        from nnsvs.pretrained import retrieve_pretrained_model
        from nnsvs.svs import SPSVS
        from nnsvs.util import example_xml_file
        import matplotlib.pyplot as plt

        # Instantiate the SVS engine
        model_dir = retrieve_pretrained_model("r9y9/yoko_latest")
        engine = SPSVS(model_dir)

        # Extract HTS labels from a MusicXML file
        contexts = pysinsy.extract_fullcontext(example_xml_file(key="get_over"))
        labels = hts.HTSLabelFile.create_from_contexts(contexts)

        # Run inference
        wav, sr = engine.svs(labels)

        # Plot the result
        fig, ax = plt.subplots(figsize=(8,2))
        librosa.display.waveshow(wav.astype(np.float32), sr=sr, ax=ax)


    With WORLD vocoder:

    >>> wav, sr = engine.svs(labels, vocoder_type="world")

    With a uSFGAN or SiFiGAN vocoder:

    >>> wav, sr = engine.svs(labels, vocoder_type="usfgan")
    """

    def __init__(self, model_dir, device="cpu", verbose=0):
        self.device = device

        # NOTE: assuming that the logger is instantiated without hydra
        # needs to add stream handler to the logger explicitly
        self.logger = getLogger(verbose=verbose, add_stream_handler=True)

        if isinstance(model_dir, str):
            model_dir = Path(model_dir)

        # search for config.yaml
        assert model_dir / "config.yaml"
        self.config = OmegaConf.load(model_dir / "config.yaml")
        self.feature_type = self.config.get("feature_type", "world")
        self.sample_rate = self.config.get("sample_rate", 48000)

        # qst
        self.binary_dict, self.numeric_dict = hts.load_question_set(
            model_dir / "qst.hed"
        )

        self.pitch_idx = get_pitch_index(self.binary_dict, self.numeric_dict)
        self.pitch_indices = get_pitch_indices(self.binary_dict, self.numeric_dict)

        # Time-lag model
        self.timelag_config = OmegaConf.load(model_dir / "timelag_model.yaml")
        self.timelag_model = instantiate(self.timelag_config.netG).to(device)
        checkpoint = torch.load(
            model_dir / "timelag_model.pth",
            map_location=device,
        )
        self.timelag_model.load_state_dict(checkpoint["state_dict"])

        self.timelag_in_scaler = MinMaxScaler(
            np.load(model_dir / "in_timelag_scaler_min.npy"),
            np.load(model_dir / "in_timelag_scaler_scale.npy"),
        )
        self.timelag_out_scaler = StandardScaler(
            np.load(model_dir / "out_timelag_scaler_mean.npy"),
            np.load(model_dir / "out_timelag_scaler_var.npy"),
            np.load(model_dir / "out_timelag_scaler_scale.npy"),
        )
        self.timelag_model.eval()

        # Duration model
        self.duration_config = OmegaConf.load(model_dir / "duration_model.yaml")
        self.duration_model = instantiate(self.duration_config.netG).to(device)
        checkpoint = torch.load(
            model_dir / "duration_model.pth",
            map_location=device,
        )
        self.duration_model.load_state_dict(checkpoint["state_dict"])

        self.duration_in_scaler = MinMaxScaler(
            np.load(model_dir / "in_duration_scaler_min.npy"),
            np.load(model_dir / "in_duration_scaler_scale.npy"),
        )
        self.duration_out_scaler = StandardScaler(
            np.load(model_dir / "out_duration_scaler_mean.npy"),
            np.load(model_dir / "out_duration_scaler_var.npy"),
            np.load(model_dir / "out_duration_scaler_scale.npy"),
        )
        self.duration_model.eval()

        # Acoustic model
        self.acoustic_config = OmegaConf.load(model_dir / "acoustic_model.yaml")
        self.acoustic_model = instantiate(self.acoustic_config.netG).to(device)
        checkpoint = torch.load(
            model_dir / "acoustic_model.pth",
            map_location=device,
        )
        self.acoustic_model.load_state_dict(checkpoint["state_dict"])
        self.acoustic_in_scaler = MinMaxScaler(
            np.load(model_dir / "in_acoustic_scaler_min.npy"),
            np.load(model_dir / "in_acoustic_scaler_scale.npy"),
        )
        self.acoustic_out_scaler = StandardScaler(
            np.load(model_dir / "out_acoustic_scaler_mean.npy"),
            np.load(model_dir / "out_acoustic_scaler_var.npy"),
            np.load(model_dir / "out_acoustic_scaler_scale.npy"),
        )
        # NOTE: this is used for GV post-filtering
        self.acoustic_out_static_scaler = extract_static_scaler(
            self.acoustic_out_scaler, self.acoustic_config
        )

        # (Optional) lf0 model
        if (model_dir / "lf0_model.pth").exists():
            assert hasattr(self.acoustic_model, "lf0_model")
            self.logger.info("Loading an external lf0 model.")
            checkpoint = torch.load(
                model_dir / "lf0_model.pth",
                map_location=device,
            )
            self.acoustic_model.lf0_model.load_state_dict(checkpoint["state_dict"])

        self.acoustic_model.eval()

        # Post-filter
        if (model_dir / "postfilter_model.yaml").exists():
            self.postfilter_config = OmegaConf.load(model_dir / "postfilter_model.yaml")
            self.postfilter_model = instantiate(self.postfilter_config.netG).to(device)
            checkpoint = torch.load(
                model_dir / "postfilter_model.pth",
                map_location=device,
            )
            self.postfilter_model.load_state_dict(checkpoint["state_dict"])
            self.postfilter_model.eval()
            self.postfilter_out_scaler = StandardScaler(
                np.load(model_dir / "out_postfilter_scaler_mean.npy"),
                np.load(model_dir / "out_postfilter_scaler_var.npy"),
                np.load(model_dir / "out_postfilter_scaler_scale.npy"),
            )
        else:
            self.postfilter_model = None
            self.postfilter_config = None
            self.postfilter_out_scaler = None

        # Vocoder model
        if (model_dir / "vocoder_model.pth").exists():
            self.vocoder, self.vocoder_in_scaler, self.vocoder_config = load_vocoder(
                model_dir / "vocoder_model.pth", device, self.acoustic_config
            )
        else:
            self.logger.info(
                "No trained vocoder model found. WORLD vocoder will be used."
            )
            self.vocoder = None
            self.vocoder_config = None
            self.vocoder_in_scaler = None

    def __repr__(self):
        timelag_str = json.dumps(
            OmegaConf.to_container(self.timelag_config.netG),
            sort_keys=False,
            indent=4,
        )
        duration_str = json.dumps(
            OmegaConf.to_container(self.duration_config.netG),
            sort_keys=False,
            indent=4,
        )
        acoustic_str = json.dumps(
            OmegaConf.to_container(self.acoustic_config.netG),
            sort_keys=False,
            indent=4,
        )

        repr = f"""Statistical parametric SVS (sampling rate: {self.sample_rate})
Time-lag model: {timelag_str}
Duration model: {duration_str}
Acoustic model: {acoustic_str}
"""
        if self.postfilter_model is not None:
            postfilter_str = json.dumps(
                OmegaConf.to_container(self.postfilter_config.netG),
                sort_keys=False,
                indent=4,
            )
            repr += f"Post-filter model: {postfilter_str}\n"
        else:
            repr += "Post-filter model: None\n"

        if self.vocoder is not None:
            if (
                "generator" in self.vocoder_config
                and "discriminator" in self.vocoder_config
            ):
                # usfgan
                vocoder_params = OmegaConf.to_container(
                    self.vocoder_config["generator"], throw_on_missing=True
                )
            else:
                vocoder_params = {
                    "generator_type": self.vocoder_config.get(
                        "generator_type", "ParallelWaveGANGenerator"  # type: ignore
                    ),
                    "generator_params": OmegaConf.to_container(
                        self.vocoder_config.generator_params
                    ),
                }
            vocoder_str = json.dumps(
                vocoder_params,
                sort_keys=False,
                indent=4,
            )
            repr += f"Vocoder model: {vocoder_str}\n"
        else:
            repr += "Vocoder model: WORLD\n"

        return repr

    def set_device(self, device):
        """Set device for the SVS model

        Args:
            device (str): cpu or cuda.
        """
        self.logger.info(f"Set device to {device}")
        self.device = device
        self.timelag_model.to(device)
        self.duration_model.to(device)
        self.acoustic_model.to(device)
        self.postfilter_model.to(device) if self.postfilter_model is not None else None
        self.vocoder.to(device) if self.vocoder is not None else None

    def predict_timelag(self, labels):
        """Predict time-ag from HTS labels

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels.

        Returns:
            ndarray: Predicted time-lag.
        """
        start_time = time.time()
        lag = predict_timelag(
            self.device,
            labels,
            timelag_model=self.timelag_model,
            timelag_config=self.timelag_config,
            timelag_in_scaler=self.timelag_in_scaler,
            timelag_out_scaler=self.timelag_out_scaler,
            binary_dict=self.binary_dict,
            numeric_dict=self.numeric_dict,
            pitch_indices=self.pitch_indices,
            log_f0_conditioning=self.config.log_f0_conditioning,
            allowed_range=self.config.timelag.allowed_range,
            allowed_range_rest=self.config.timelag.allowed_range_rest,
            force_clip_input_features=self.config.timelag.force_clip_input_features,
            frame_period=self.config.frame_period,
        )
        self.logger.info(
            f"Elapsed time for time-lag prediction: {time.time() - start_time:.3f} sec"
        )
        return lag

    def predict_duration(self, labels):
        """Predict durations from HTS labels

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels.

        Returns:
            ndarray: Predicted durations.
        """
        start_time = time.time()
        durations = predict_duration(
            self.device,
            labels,
            duration_model=self.duration_model,
            duration_config=self.duration_config,
            duration_in_scaler=self.duration_in_scaler,
            duration_out_scaler=self.duration_out_scaler,
            binary_dict=self.binary_dict,
            numeric_dict=self.numeric_dict,
            pitch_indices=self.pitch_indices,
            log_f0_conditioning=self.config.log_f0_conditioning,
            force_clip_input_features=self.config.duration.force_clip_input_features,
            frame_period=self.config.frame_period,
        )
        self.logger.info(
            f"Elapsed time for duration prediction: {time.time() - start_time:.3f} sec"
        )
        return durations

    def postprocess_duration(self, labels, pred_durations, lag):
        """Post-process durations

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels.
            pred_durations (ndarray): Predicted durations.
            lag (ndarray): Predicted time-lag.

        Returns:
            nnmnkwii.io.hts.HTSLabelFile: duration modified HTS labels.
        """
        start_time = time.time()
        duration_modified_labels = postprocess_duration(
            labels, pred_durations, lag, frame_period=self.config.frame_period
        )
        self.logger.info(
            f"Elapsed time for duration post-processing: {time.time() - start_time:.3f} sec"
        )
        return duration_modified_labels

    def predict_timing(self, labels):
        """Predict timing from HTS labels

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels.

        Returns:
            nnmnkwii.io.hts.HTSLabelFile: duration modified HTS labels.
        """
        lag = self.predict_timelag(labels)
        durations = self.predict_duration(labels)
        duration_modified_full_labels = self.postprocess_duration(
            labels, durations, lag
        )
        return duration_modified_full_labels

    def predict_acoustic(self, duration_modified_labels, f0_shift_in_cent=0):
        """Predict acoustic features from HTS labels

        Args:
            duration_modified_labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels.
            f0_shift_in_cent (float): F0 shift in cent.

        Returns:
            ndarray: Predicted acoustic features.
        """
        start_time = time.time()
        acoustic_features = predict_acoustic(
            device=self.device,
            labels=duration_modified_labels,
            acoustic_model=self.acoustic_model,
            acoustic_config=self.acoustic_config,
            acoustic_in_scaler=self.acoustic_in_scaler,
            acoustic_out_scaler=self.acoustic_out_scaler,
            binary_dict=self.binary_dict,
            numeric_dict=self.numeric_dict,
            subphone_features=self.acoustic_config.get(
                "subphone_features", "coarse_coding"
            ),
            pitch_indices=self.pitch_indices,
            log_f0_conditioning=self.config.log_f0_conditioning,
            force_clip_input_features=self.acoustic_config.get(
                "force_clip_input_features", True
            ),
            frame_period=self.config.frame_period,
            f0_shift_in_cent=f0_shift_in_cent,
        )
        self.logger.info(
            f"Elapsed time for acoustic feature prediction: {time.time() - start_time:.3f} sec"
        )
        # log real-time factor (RT)
        RT = (time.time() - start_time) / (
            acoustic_features.shape[0] * self.config.frame_period / 1000
        )
        self.logger.info(f"Real-time factor for acoustic feature prediction: {RT:.3f}")
        return acoustic_features

    def postprocess_acoustic(
        self,
        duration_modified_labels,
        acoustic_features,
        post_filter_type="gv",
        trajectory_smoothing=True,
        trajectory_smoothing_cutoff=50,
        trajectory_smoothing_cutoff_f0=20,
        vuv_threshold=0.5,
        force_fix_vuv=False,
        fill_silence_to_rest=False,
        f0_shift_in_cent=0,
    ):
        """Post-process acoustic features

        The function converts acoustic features in single ndarray to tuple of
        multi-stream acoustic features.

        e.g., array -> (mgc, lf0, vuv, bap)

        If post_filter_type=``nnsvs`` is specified, learned post-filter is applied.
        However, it is recommended to use ``gv`` in general.

        Args:
            duration_modified_labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels.
            acoustic_features (ndarray): Predicted acoustic features.
            post_filter_type (str): Post-filter type.
                One of ``gv``, ``merlin`` or ``nnsvs``. Recommended to use ``gv``
                for general purpose.
            trajectory_smoothing (bool): Whether to apply trajectory smoothing.
            trajectory_smoothing_cutoff (float): Cutoff frequency for trajectory smoothing
                of spectral features.
            trajectory_smoothing_cutoff_f0 (float): Cutoff frequency for trajectory
                smoothing of f0.
            vuv_threshold (float): V/UV threshold.
            force_fix_vuv (bool): Force fix V/UV.
            fill_silence_to_rest (bool): Fill silence to rest frames.
            f0_shift_in_cent (float): F0 shift in cent.

        Returns:
            tuple: Post-processed multi-stream acoustic features.
        """
        start_time = time.time()
        multistream_features = postprocess_acoustic(
            device=self.device,
            duration_modified_labels=duration_modified_labels,
            acoustic_features=acoustic_features,
            binary_dict=self.binary_dict,
            numeric_dict=self.numeric_dict,
            acoustic_config=self.acoustic_config,
            acoustic_out_static_scaler=self.acoustic_out_static_scaler,
            postfilter_model=self.postfilter_model,
            postfilter_config=self.postfilter_config,
            postfilter_out_scaler=self.postfilter_out_scaler,
            sample_rate=self.sample_rate,
            frame_period=self.config.frame_period,
            relative_f0=self.config.acoustic.relative_f0,
            feature_type=self.feature_type,
            post_filter_type=post_filter_type,
            trajectory_smoothing=trajectory_smoothing,
            trajectory_smoothing_cutoff=trajectory_smoothing_cutoff,
            trajectory_smoothing_cutoff_f0=trajectory_smoothing_cutoff_f0,
            vuv_threshold=vuv_threshold,
            f0_shift_in_cent=f0_shift_in_cent,
            vibrato_scale=1.0,  # only valid for Sinsy-like models
            force_fix_vuv=force_fix_vuv,
            fill_silence_to_rest=fill_silence_to_rest,
        )
        self.logger.info(
            f"Elapsed time for acoustic post-processing: {time.time() - start_time:.3f} sec"
        )
        return multistream_features

    def predict_waveform(
        self,
        multistream_features,
        vocoder_type="world",
        vuv_threshold=0.5,
    ):
        """Predict waveform from acoustic features

        Args:
            multistream_features (tuple): Multi-stream acoustic features.
            vocoder_type (str): Vocoder type. One of ``world``, ``pwg`` or ``usfgan``.
                If ``auto`` is specified, the vocoder is automatically selected.
            vuv_threshold (float): V/UV threshold.

        Returns:
            ndarray: Predicted waveform.
        """
        start_time = time.time()
        if vocoder_type in ["pwg", "usfgan"] and self.vocoder is None:
            raise ValueError(
                """Pre-trained vocodr model is not found.
WORLD is only supported for waveform generation"""
            )

        if vocoder_type == "auto":
            if self.feature_type == "melf0":
                assert self.vocoder is not None
                vocoder_type = (
                    "usfgan" if isinstance(self.vocoder, USFGANWrapper) else "pwg"
                )
            elif self.feature_type == "world":
                if self.vocoder is None:
                    vocoder_type = "world"
                else:
                    vocoder_type = (
                        "usfgan" if isinstance(self.vocoder, USFGANWrapper) else "pwg"
                    )
        wav = predict_waveform(
            device=self.device,
            multistream_features=multistream_features,
            vocoder=self.vocoder,
            vocoder_config=self.vocoder_config,
            vocoder_in_scaler=self.vocoder_in_scaler,
            sample_rate=self.sample_rate,
            frame_period=self.config.frame_period,
            use_world_codec=self.config.get("use_world_codec", False),
            feature_type=self.feature_type,
            vocoder_type=vocoder_type,
            vuv_threshold=vuv_threshold,
        )
        self.logger.info(
            f"Elapsed time for waveform generation: {time.time() - start_time:.3f} sec"
        )
        RT = (time.time() - start_time) / (len(wav) / self.sample_rate)
        self.logger.info(f"Real-time factor for waveform generation: {RT:.3f}")
        return wav

    def postprocess_waveform(
        self,
        wav,
        dtype=np.int16,
        peak_norm=False,
        loudness_norm=False,
        target_loudness=-20,
    ):
        """Post-process waveform

        Args:
            wav (ndarray): Waveform.
            dtype (dtype): Data type of waveform.
            peak_norm (bool): Whether to apply peak normalization.
            loudness_norm (bool): Whether to apply loudness normalization.
            target_loudness (float): Target loudness in dB.

        Returns:
            ndarray: Post-processed waveform.
        """
        start_time = time.time()
        wav = postprocess_waveform(
            wav=wav,
            sample_rate=self.sample_rate,
            dtype=dtype,
            peak_norm=peak_norm,
            loudness_norm=loudness_norm,
            target_loudness=target_loudness,
        )
        self.logger.info(
            f"Elapsed time for waveform post-processing: {time.time() - start_time:.3f} sec"
        )
        return wav

    def svs(
        self,
        labels,
        vocoder_type="world",
        post_filter_type="gv",
        trajectory_smoothing=True,
        trajectory_smoothing_cutoff=50,
        trajectory_smoothing_cutoff_f0=20,
        vuv_threshold=0.5,
        style_shift=0,
        force_fix_vuv=False,
        fill_silence_to_rest=False,
        dtype=np.int16,
        peak_norm=False,
        loudness_norm=False,
        target_loudness=-20,
        segmented_synthesis=False,
    ):
        """Synthesize waveform from HTS labels.

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels
            vocoder_type (str): Vocoder type. One of ``world``, ``pwg`` or ``usfgan``.
                If ``auto`` is specified, the vocoder is automatically selected.
            post_filter_type (str): Post-filter type. ``merlin``, ``gv`` or ``nnsvs``
                is supported.
            trajectory_smoothing (bool): Whether to smooth acoustic feature trajectory.
            trajectory_smoothing_cutoff (int): Cutoff frequency for trajectory smoothing.
            trajectory_smoothing_cutoff_f0 (int): Cutoff frequency for trajectory
                smoothing of f0.
            vuv_threshold (float): Threshold for VUV.
            style_shift (int): style shift parameter
            force_fix_vuv (bool): Whether to correct VUV.
            fill_silence_to_rest (bool): Fill silence to rest frames.
            dtype (np.dtype): Data type of the output waveform.
            peak_norm (bool): Whether to normalize the waveform by peak value.
            loudness_norm (bool): Whether to normalize the waveform by loudness.
            target_loudness (float): Target loudness in dB.
            segmneted_synthesis (bool): Whether to use segmented synthesis.
        """
        start_time = time.time()
        vocoder_type = vocoder_type.lower()
        if vocoder_type not in ["world", "pwg", "usfgan", "auto"]:
            raise ValueError(f"Unknown vocoder type: {vocoder_type}")
        if post_filter_type not in ["merlin", "nnsvs", "gv", "none"]:
            raise ValueError(f"Unknown post-filter type: {post_filter_type}")

        # Predict timinigs
        duration_modified_labels = self.predict_timing(labels)

        # NOTE: segmented synthesis is not well tested. There MUST be better ways
        # to do this.
        if segmented_synthesis:
            self.logger.warning(
                "Segmented synthesis is not well tested. Use it on your own risk."
            )
            duration_modified_labels_segs = segment_labels(
                duration_modified_labels,
                # the following parameters are based on experiments in the NNSVS's paper
                # tuned with Namine Ritsu's database
                silence_threshold=0.1,
                min_duration=5.0,
                force_split_threshold=5.0,
            )
            from tqdm.auto import tqdm
        else:
            duration_modified_labels_segs = [duration_modified_labels]

            def tqdm(x, **kwargs):
                return x

        # Run acoustic model and vocoder
        hts_frame_shift = int(self.config.frame_period * 1e4)
        wavs = []
        self.logger.info(f"Number of segments: {len(duration_modified_labels_segs)}")
        for duration_modified_labels_seg in tqdm(
            duration_modified_labels_segs,
            desc="[segment]",
            total=len(duration_modified_labels_segs),
        ):
            duration_modified_labels_seg.frame_shift = hts_frame_shift

            # Predict acoustic features
            # NOTE: if non-zero pre_f0_shift_in_cent is specified, the input pitch
            # will be shifted before running the acoustic model
            acoustic_features = self.predict_acoustic(
                duration_modified_labels_seg,
                f0_shift_in_cent=style_shift * 100,
            )

            # Post-processing for acoustic features
            # NOTE: if non-zero post_f0_shift_in_cent is specified, the output pitch
            # will be shifted as a part of post-processing
            multistream_features = self.postprocess_acoustic(
                acoustic_features=acoustic_features,
                duration_modified_labels=duration_modified_labels_seg,
                trajectory_smoothing=trajectory_smoothing,
                trajectory_smoothing_cutoff=trajectory_smoothing_cutoff,
                trajectory_smoothing_cutoff_f0=trajectory_smoothing_cutoff_f0,
                force_fix_vuv=force_fix_vuv,
                fill_silence_to_rest=fill_silence_to_rest,
                f0_shift_in_cent=-style_shift * 100,
            )

            # Generate waveform by vocoder
            wav = self.predict_waveform(
                multistream_features=multistream_features,
                vocoder_type=vocoder_type,
                vuv_threshold=vuv_threshold,
            )

            wavs.append(wav)

        # Concatenate segmented waveforms
        wav = np.concatenate(wavs, axis=0).reshape(-1)

        # Post-processing for the output waveform
        wav = self.postprocess_waveform(
            wav,
            dtype=dtype,
            peak_norm=peak_norm,
            loudness_norm=loudness_norm,
            target_loudness=target_loudness,
        )
        self.logger.info(f"Total time: {time.time() - start_time:.3f} sec")
        RT = (time.time() - start_time) / (len(wav) / self.sample_rate)
        self.logger.info(f"Total real-time factor: {RT:.3f}")
        return wav, self.sample_rate


def _warn_if_model_is_old(logger):
    logger.warning(
        """It is likely you have trained you model with old NNSVS.
It is recommended to retrain your model with the latest version of NNSVS."""
    )


class NEUTRINO(SPSVS):
    """NEUTRINO-like interface for singing voice synthesis

    Args:
        model_dir (str): model directory
        device (str): device name
        verbose (int): verbose level
    """

    def __init__(self, model_dir, device="cpu", verbose=0):
        super().__init__(model_dir, device=device, verbose=verbose)

        if self.feature_type != "world":
            raise RuntimeError(f"Unsupported feature type: {self.feature_type}")
        if not self.config.get("use_world_codec", False):
            self.logger.warning(
                "WORLD coded is required to output NEUTRIN-compatible features"
            )
            _warn_if_model_is_old(self.logger)

    @classmethod
    def musicxml2label(cls, input_file):
        """Convert musicXML to full and mono HTS labels

        Args:
            input_file (str): musicXML file
        """
        import pysinsy

        contexts = pysinsy.extract_fullcontext(input_file)
        full_labels = hts.HTSLabelFile.create_from_contexts(contexts)
        mono_labels = full_to_mono(full_labels)

        return full_labels, mono_labels

    def get_num_phrases(self, labels):
        """Get number of phrases

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS label

        Returns:
            int: number of phrases
        """
        phrases = label2phrases(labels)
        return len(phrases)

    def get_phraselist(self, full_labels, timing_labels):
        """Get phraselit from full and timing HTS labels

        Args:
            full_labels (nnmnkwii.io.hts.HTSLabelFile): full HTS label
            timing_labels (nnmnkwii.io.hts.HTSLabelFile): timing HTS label

        Returns:
            str: phraselist
        """
        note_indices = get_note_indices(full_labels)
        phraselist = label2phrases_str(timing_labels, note_indices)
        return phraselist

    def predict_acoustic(
        self,
        full_labels,
        timing_labels=None,
        style_shift=0,
        phrase_num=-1,
        trajectory_smoothing=True,
        trajectory_smoothing_cutoff=50,
        trajectory_smoothing_cutoff_f0=20,
        vuv_threshold=0.5,
        force_fix_vuv=False,
        fill_silence_to_rest=False,
    ):
        """Main inference of timing and acoustic predictions

        Args:
            full_labels (nnmnkwii.io.hts.HTSLabelFile): full HTS label
            timing_labels (nnmnkwii.io.hts.HTSLabelFile): timing HTS label
            style_shift (int): style shift parameter
            phrase_num (int): phrase number to use for inference
            trajectory_smoothing (bool): whether to apply trajectory smoothing
            trajectory_smoothing_cutoff (float): cutoff frequency for trajectory smoothing
            trajectory_smoothing_cutoff_f0 (float): cutoff frequency for trajectory
                smoothing for f0
            vuv_threshold (float): V/UV threshold
            force_fix_vuv (bool): whether to force fix V/UV
            fill_silence_to_rest (bool): Fill silence to rest frames.

        Returns:
            tuple: (f0, mgc, bap)
        """
        if timing_labels is None:
            self.logger.warning("'timing_labels' is not provided.")
            # Run timing prediction
            duration_modified_full_labels = self.predict_timing(full_labels)
            timing_labels = full_to_mono(duration_modified_full_labels)
        else:
            # Load pre-estimated timing
            duration_modified_full_labels = deepcopy(full_labels)
            duration_modified_full_labels.start_times = timing_labels.start_times.copy()
            duration_modified_full_labels.end_times = timing_labels.end_times.copy()

        if phrase_num >= 0:
            phrases = label2phrases(duration_modified_full_labels)
            if phrase_num > len(phrases):
                raise RuntimeError(
                    f"phrase_num is too large: {phrase_num} > {len(phrases)}"
                )

            # Use the specified phrase for inference
            duration_modified_full_labels = phrases[phrase_num]
            self.logger.info(f"Using phrase {phrase_num}/{len(phrases)} for inference")

        # Predict acoustic features
        # NOTE: if non-zero pre_f0_shift_in_cent is specified, the input pitch
        # will be shifted before running the acoustic model
        acoustic_features = super().predict_acoustic(
            duration_modified_full_labels,
            f0_shift_in_cent=style_shift * 100,
        )

        # Post-processing for acoustic features
        # NOTE: if non-zero post_f0_shift_in_cent is specified, the output pitch
        # will be shifted as a part of post-processing
        multistream_features = super().postprocess_acoustic(
            acoustic_features=acoustic_features,
            duration_modified_labels=duration_modified_full_labels,
            trajectory_smoothing=trajectory_smoothing,
            trajectory_smoothing_cutoff=trajectory_smoothing_cutoff,
            trajectory_smoothing_cutoff_f0=trajectory_smoothing_cutoff_f0,
            vuv_threshold=vuv_threshold,
            force_fix_vuv=force_fix_vuv,
            fill_silence_to_rest=fill_silence_to_rest,
            f0_shift_in_cent=-style_shift * 100,
        )
        assert len(multistream_features) == 4
        mgc, lf0, vuv, bap = multistream_features

        if not self.config.get("use_world_codec", False):
            self.logger.warning(
                """use_world_codec is not set.
 WORLD (NEUTRINO edition) does not work with the output of this model.
            """
            )

        # Convert lf0 to f0
        f0 = np.exp(lf0.copy())
        f0[vuv < vuv_threshold] = 0

        # NOTE: Neutrino-compatible MGC should have negative values at the 0-th coefficient.
        if mgc[:, 0].mean() > 0:
            self.logger.warning("MGC 0-th coefficient is positive.")
            _warn_if_model_is_old(self.logger)

        # Make sure to have correct array layout and dtype
        # These parameters can be used to generate waveform by WORLD
        f0 = np.ascontiguousarray(f0).astype(np.float64)
        mgc = np.ascontiguousarray(mgc).astype(np.float64)
        bap = np.ascontiguousarray(bap).astype(np.float64)
        return f0, mgc, bap

    def predict_waveform(
        self,
        f0,
        mgc,
        bap,
        vocoder_type="world",
        vuv_threshold=0.5,
        dtype=np.int16,
        peak_norm=False,
        loudness_norm=False,
        target_loudness=-20,
    ):
        """Generate waveform from acoustic features

        Args:
            f0 (ndarray): f0
            mgc (ndarray): mel-cepstrum
            bap (ndarray): band-aperiodicity
            vocoder_type (str): vocoder type
            vuv_threshold (float): V/UV threshold
            dtype (np.dtype): Data type of the output waveform.
            peak_norm (bool): Whether to normalize the waveform by peak value.
            loudness_norm (bool): Whether to normalize the waveform by loudness.
            target_loudness (float): Target loudness in dB.

        Returns:
            ndarray: waveform
        """
        # Convert NEUTRINO-like features to NNSVS's one
        # (f0, mgc, bap) -> (mgc, lf0, vuv, bap)
        vuv = (f0 > 0).astype(np.float64).reshape(-1, 1)
        lf0 = f0.copy()
        lf0[np.nonzero(lf0)] = np.log(f0[np.nonzero(lf0)])
        lf0 = interp1d(lf0, kind="slinear")
        multistream_features = (mgc, lf0, vuv, bap)

        wav = super().predict_waveform(
            multistream_features=multistream_features,
            vocoder_type=vocoder_type,
            vuv_threshold=vuv_threshold,
        )
        wav = self.postprocess_waveform(
            wav,
            dtype=dtype,
            peak_norm=peak_norm,
            loudness_norm=loudness_norm,
            target_loudness=target_loudness,
        )
        return wav

    def svs(self, labels):
        """Synthesize wavefrom from HTS labels

        Args:
            labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels

        Returns:
            tuple: (waveform, sample_rate)
        """
        self.logger.warning(
            "Use `predict_acoustic` and `predict_waveform` methods instead."
        )
        f0, mgc, bap = self.predict_acoustic(labels)
        wav = self.predict_waveform(f0, mgc, bap)
        return wav, self.sample_rate



import numpy as np
import torch
from nnsvs.acoustic_models.util import pad_inference
from nnsvs.base import BaseModel, PredictionType
from nnsvs.mdn import MDNLayer, mdn_get_most_probable_sigma_and_mu, mdn_get_sample
from nnsvs.tacotron.decoder import Prenet, ZoneOutCell
from nnsvs.util import init_weights
from torch import nn
from torch.nn import functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

__all__ = [
    "ResF0NonAttentiveDecoder",
    "MDNResF0NonAttentiveDecoder",
    "BiLSTMResF0NonAttentiveDecoder",
]


class ResF0NonAttentiveDecoder(BaseModel):
    """Duration-informed Tacotron with residual F0 prediction.

    Args:
        in_dim (int) : dimension of encoder hidden layer
        out_dim (int) : dimension of output
        layers (int) : number of LSTM layers
        hidden_dim (int) : dimension of hidden layer
        prenet_layers (int) : number of pre-net layers
        prenet_hidden_dim (int) : dimension of pre-net hidden layer
        prenet_dropout (float) : dropout rate of pre-net
        zoneout (float) : zoneout rate
        reduction_factor (int) : reduction factor
        downsample_by_conv (bool) : if True, downsample by convolution
        scaled_tanh (bool) : if True, use scaled tanh for residual F0 prediction
        in_lf0_idx (int): index of lf0 in input features
        in_lf0_min (float): minimum of lf0 in the training data of input features
        in_lf0_max (float): maximum of lf0 in the training data of input features
        out_lf0_idx (int): index of lf0 in output features
        out_lf0_mean (float): mean of lf0 in the training data of output features
        out_lf0_scale (float): scale of lf0 in the training data of output features
        init_type (str): initialization type
        eval_dropout (bool): if True, use dropout in evaluation
    """

    def __init__(
        self,
        in_dim=512,
        out_dim=1,
        layers=2,
        hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        scaled_tanh=True,
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
        init_type="none",
        eval_dropout=True,
    ):
        super().__init__()
        self.out_dim = out_dim
        self.reduction_factor = reduction_factor
        self.prenet_dropout = prenet_dropout
        self.scaled_tanh = scaled_tanh
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale

        if prenet_layers > 0:
            self.prenet = Prenet(
                out_dim,
                prenet_layers,
                prenet_hidden_dim,
                prenet_dropout,
                eval_dropout=eval_dropout,
            )
            lstm_in_dim = in_dim + prenet_hidden_dim
        else:
            self.prenet = None
            prenet_hidden_dim = 0
            lstm_in_dim = in_dim + out_dim

        self.lstm = nn.ModuleList()
        for layer in range(layers):
            lstm = nn.LSTMCell(
                lstm_in_dim if layer == 0 else hidden_dim,
                hidden_dim,
            )
            self.lstm += [ZoneOutCell(lstm, zoneout)]

        proj_in_dim = in_dim + hidden_dim
        self.feat_out = nn.Linear(proj_in_dim, out_dim * reduction_factor, bias=False)

        if reduction_factor > 1 and downsample_by_conv:
            self.conv_downsample = nn.Conv1d(
                in_dim,
                in_dim,
                kernel_size=reduction_factor,
                stride=reduction_factor,
                groups=in_dim,
            )
        else:
            self.conv_downsample = None
        init_weights(self, init_type)

    def _zero_state(self, hs):
        init_hs = hs.new_zeros(hs.size(0), self.lstm[0].hidden_size)
        return init_hs

    def is_autoregressive(self):
        return True

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, encoder_outs, in_lens, decoder_targets=None):
        """Forward step

        Args:
            encoder_outs (torch.Tensor): encoder outputs (B, T, C)
            in_lens (torch.Tensor): input lengths
            decoder_targets (torch.Tensor): decoder targets for teacher-forcing. (B, T, C)

        Returns:
            torch.Tensor: the output (B, C, T)
        """
        is_inference = decoder_targets is None
        if not is_inference:
            assert encoder_outs.shape[1] == decoder_targets.shape[1]

        # Denormalize lf0 from input musical score
        lf0_score = encoder_outs[:, :, self.in_lf0_idx].unsqueeze(-1)
        lf0_score_denorm = (
            lf0_score * (self.in_lf0_max - self.in_lf0_min) + self.in_lf0_min
        )
        # (B, T, C) -> (B, C, T)
        lf0_score_denorm = lf0_score_denorm.transpose(1, 2)

        # To avoid unbounded residual f0 that would potentially cause artifacts,
        # let's constrain the residual F0 to be in a certain range by the scaled tanh
        residual_f0_max_cent = 600
        max_lf0_ratio = residual_f0_max_cent * np.log(2) / 1200

        # Adjust number of frames according to the reduction factor
        # (B, Lmax, out_dim) ->  (B, Lmax/r, out_dim)
        if self.reduction_factor > 1 and not is_inference:
            decoder_targets = decoder_targets[
                :, self.reduction_factor - 1 :: self.reduction_factor
            ]
        if self.reduction_factor > 1:
            if self.conv_downsample is not None:
                encoder_outs = self.conv_downsample(
                    encoder_outs.transpose(1, 2)
                ).transpose(1, 2)
            else:
                encoder_outs = encoder_outs[
                    :, self.reduction_factor - 1 :: self.reduction_factor
                ]

        h_list, c_list = [], []
        for _ in range(len(self.lstm)):
            h_list.append(self._zero_state(encoder_outs))
            c_list.append(self._zero_state(encoder_outs))

        go_frame = encoder_outs.new_zeros(encoder_outs.size(0), self.out_dim)
        prev_out = go_frame

        if not is_inference and self.prenet is not None:
            prenet_outs = self.prenet(decoder_targets)

        outs = []
        lf0_residuals = []
        for t in range(encoder_outs.shape[1]):
            # Pre-Net
            if self.prenet is not None:
                if is_inference:
                    prenet_out = self.prenet(prev_out)
                else:
                    prenet_out = prenet_outs[:, t, :]
            else:
                prenet_out = F.dropout(prev_out, self.prenet_dropout, training=True)

            # LSTM
            xs = torch.cat([encoder_outs[:, t], prenet_out], dim=1)
            h_list[0], c_list[0] = self.lstm[0](xs, (h_list[0], c_list[0]))
            for i in range(1, len(self.lstm)):
                h_list[i], c_list[i] = self.lstm[i](
                    h_list[i - 1], (h_list[i], c_list[i])
                )
            # Output
            hcs = torch.cat([h_list[-1], encoder_outs[:, t]], dim=1)
            out = self.feat_out(hcs).view(encoder_outs.size(0), self.out_dim, -1)
            # Residual F0
            if self.scaled_tanh:
                lf0_residual = max_lf0_ratio * torch.tanh(
                    out[:, self.out_lf0_idx, :]
                ).unsqueeze(1)
            else:
                lf0_residual = out[:, self.out_lf0_idx, :].unsqueeze(1)

            # Residual connection in the denormalized f0 domain
            lf0_score_denorm_t = lf0_score_denorm[
                :, :, t * self.reduction_factor : (t + 1) * self.reduction_factor
            ]
            lf0_pred_denorm = lf0_score_denorm_t + lf0_residual
            # Back to normalized f0
            lf0_pred = (lf0_pred_denorm - self.out_lf0_mean) / self.out_lf0_scale
            out[:, self.out_lf0_idx, :] = lf0_pred.squeeze(1)

            outs.append(out)
            lf0_residuals.append(lf0_residual)

            # Update decoder input for the next time step
            if is_inference:
                prev_out = outs[-1][:, :, -1]  # (1, out_dim)
            else:
                # Teacher forcing
                prev_out = decoder_targets[:, t, :]

        outs = torch.cat(outs, dim=2)  # (B, out_dim, Lmax)
        lf0_residuals = torch.cat(lf0_residuals, dim=2)  # (B, out_dim, Lmax)

        # (B, C, T) -> (B, T, C)
        outs = outs.transpose(1, 2)
        lf0_residuals = lf0_residuals.transpose(1, 2)

        return outs, lf0_residuals

    def inference(self, x, lengths=None):
        return pad_inference(
            model=self, x=x, lengths=lengths, reduction_factor=self.reduction_factor
        )


class MDNResF0NonAttentiveDecoder(BaseModel):
    """Duration-informed Tacotron with residual F0 prediction (MDN-version)

    Args:
        in_dim (int) : dimension of encoder hidden layer
        out_dim (int) : dimension of output
        layers (int) : number of LSTM layers
        hidden_dim (int) : dimension of hidden layer
        prenet_layers (int) : number of pre-net layers
        prenet_hidden_dim (int) : dimension of pre-net hidden layer
        prenet_dropout (float) : dropout rate of pre-net
        zoneout (float) : zoneout rate
        reduction_factor (int) : reduction factor
        downsample_by_conv (bool) : if True, downsample by convolution
        scaled_tanh (bool) : if True, use scaled tanh for residual F0 prediction
        num_gaussians (int) : number of Gaussian
        sampling_mode (str) : sampling mode
        in_lf0_idx (int): index of lf0 in input features
        in_lf0_min (float): minimum of lf0 in the training data of input features
        in_lf0_max (float): maximum of lf0 in the training data of input features
        out_lf0_idx (int): index of lf0 in output features
        out_lf0_mean (float): mean of lf0 in the training data of output features
        out_lf0_scale (float): scale of lf0 in the training data of output features
        init_type (str): initialization type
        eval_dropout (bool): if True, use dropout in evaluation
    """

    def __init__(
        self,
        in_dim=512,
        out_dim=80,
        layers=2,
        hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        scaled_tanh=True,
        num_gaussians=4,
        sampling_mode="mean",
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
        init_type="none",
        eval_dropout=True,
    ):
        super().__init__()
        self.out_dim = out_dim
        self.reduction_factor = reduction_factor
        self.prenet_dropout = prenet_dropout
        self.scaled_tanh = scaled_tanh
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale
        self.num_gaussians = num_gaussians
        self.sampling_mode = sampling_mode
        assert sampling_mode in ["mean", "random"]

        if prenet_layers > 0:
            self.prenet = Prenet(
                out_dim,
                prenet_layers,
                prenet_hidden_dim,
                prenet_dropout,
                eval_dropout=eval_dropout,
            )
            lstm_in_dim = in_dim + prenet_hidden_dim
        else:
            self.prenet = None
            prenet_hidden_dim = 0
            lstm_in_dim = in_dim + out_dim

        self.lstm = nn.ModuleList()
        for layer in range(layers):
            lstm = nn.LSTMCell(
                lstm_in_dim if layer == 0 else hidden_dim,
                hidden_dim,
            )
            self.lstm += [ZoneOutCell(lstm, zoneout)]

        proj_in_dim = in_dim + hidden_dim
        self.feat_out = MDNLayer(
            proj_in_dim,
            out_dim * reduction_factor,
            num_gaussians=num_gaussians,
            dim_wise=True,
        )

        if reduction_factor > 1 and downsample_by_conv:
            self.conv_downsample = nn.Conv1d(
                in_dim,
                in_dim,
                kernel_size=reduction_factor,
                stride=reduction_factor,
                groups=in_dim,
            )
        else:
            self.conv_downsample = None
        init_weights(self, init_type)

    def _zero_state(self, hs):
        init_hs = hs.new_zeros(hs.size(0), self.lstm[0].hidden_size)
        return init_hs

    def prediction_type(self):
        return PredictionType.PROBABILISTIC

    def is_autoregressive(self):
        return True

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, encoder_outs, in_lens, decoder_targets=None):
        """Forward step

        Args:
            encoder_outs (torch.Tensor): encoder outputs (B, T, C)
            in_lens (torch.Tensor): input lengths
            decoder_targets (torch.Tensor): decoder targets for teacher-forcing. (B, T, C)

        Returns:
            torch.Tensor: the output (B, C, T)
        """
        is_inference = decoder_targets is None
        if not is_inference:
            assert encoder_outs.shape[1] == decoder_targets.shape[1]

        # Denormalize lf0 from input musical score
        lf0_score = encoder_outs[:, :, self.in_lf0_idx].unsqueeze(-1)
        lf0_score_denorm = (
            lf0_score * (self.in_lf0_max - self.in_lf0_min) + self.in_lf0_min
        )
        # (B, T, C) -> (B, C, T)
        # (B, T, C)
        # lf0_score_denorm = lf0_score_denorm.transpose(1, 2)

        # To avoid unbounded residual f0 that would potentially cause artifacts,
        # let's constrain the residual F0 to be in a certain range by the scaled tanh
        residual_f0_max_cent = 600
        max_lf0_ratio = residual_f0_max_cent * np.log(2) / 1200

        # Adjust number of frames according to the reduction factor
        # (B, Lmax, out_dim) ->  (B, Lmax/r, out_dim)
        if self.reduction_factor > 1 and not is_inference:
            decoder_targets = decoder_targets[
                :, self.reduction_factor - 1 :: self.reduction_factor
            ]
        if self.reduction_factor > 1:
            if self.conv_downsample is not None:
                encoder_outs = self.conv_downsample(
                    encoder_outs.transpose(1, 2)
                ).transpose(1, 2)
            else:
                encoder_outs = encoder_outs[
                    :, self.reduction_factor - 1 :: self.reduction_factor
                ]

        h_list, c_list = [], []
        for _ in range(len(self.lstm)):
            h_list.append(self._zero_state(encoder_outs))
            c_list.append(self._zero_state(encoder_outs))

        go_frame = encoder_outs.new_zeros(encoder_outs.size(0), self.out_dim)
        prev_out = go_frame

        if not is_inference and self.prenet is not None:
            prenet_outs = self.prenet(decoder_targets)

        mus = []
        log_pis = []
        log_sigmas = []
        lf0_residuals = []
        mus_inf = []
        for t in range(encoder_outs.shape[1]):
            # Pre-Net
            if self.prenet is not None:
                if is_inference:
                    prenet_out = self.prenet(prev_out)
                else:
                    prenet_out = prenet_outs[:, t, :]
            else:
                prenet_out = F.dropout(prev_out, self.prenet_dropout, training=True)

            # LSTM
            xs = torch.cat([encoder_outs[:, t], prenet_out], dim=1)
            h_list[0], c_list[0] = self.lstm[0](xs, (h_list[0], c_list[0]))
            for i in range(1, len(self.lstm)):
                h_list[i], c_list[i] = self.lstm[i](
                    h_list[i - 1], (h_list[i], c_list[i])
                )
            # Output
            hcs = torch.cat([h_list[-1], encoder_outs[:, t]], dim=1)
            log_pi, log_sigma, mu = self.feat_out(hcs.unsqueeze(1))

            # (B, reduction_factor, num_gaussians, out_dim)
            log_pi = (
                log_pi.transpose(1, 2)
                .view(encoder_outs.size(0), self.num_gaussians, -1, self.out_dim)
                .transpose(1, 2)
            )
            log_sigma = (
                log_sigma.transpose(1, 2)
                .view(encoder_outs.size(0), self.num_gaussians, -1, self.out_dim)
                .transpose(1, 2)
            )
            mu = (
                mu.transpose(1, 2)
                .view(encoder_outs.size(0), self.num_gaussians, -1, self.out_dim)
                .transpose(1, 2)
            )

            # Residual F0
            # (B, reduction_factor, num_gaussians)
            if self.scaled_tanh:
                lf0_residual = max_lf0_ratio * torch.tanh(mu[:, :, :, self.out_lf0_idx])
            else:
                lf0_residual = mu[:, :, :, self.out_lf0_idx]

            # Residual connection in the denormalized f0 domain
            lf0_score_denorm_t = lf0_score_denorm[
                :, t * self.reduction_factor : (t + 1) * self.reduction_factor, :
            ]
            # NOTE: broadcast against num_gaussians axis
            # (B, 1, 1) + (B, 1, num_gaussians) -> (B, 1, num_gaussians)
            lf0_pred_denorm = lf0_score_denorm_t + lf0_residual
            # Back to normalized f0
            lf0_pred = (lf0_pred_denorm - self.out_lf0_mean) / self.out_lf0_scale
            mu[:, :, :, self.out_lf0_idx] = lf0_pred

            mus.append(mu)
            log_pis.append(log_pi)
            log_sigmas.append(log_sigma)
            lf0_residuals.append(lf0_residual)

            # Update decoder input for the next time step
            if is_inference:
                # (B, reduction_factor, out_dim)
                if self.sampling_mode == "mean":
                    _, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
                elif self.sampling_mode == "random":
                    mu = mdn_get_sample(log_pi, log_sigma, mu)
                # Feed last sample for the feedback loop
                prev_out = mu[:, -1]

                mus_inf.append(mu)
            else:
                # Teacher forcing
                prev_out = decoder_targets[:, t, :]

        # (B, T, G, out_dim)
        mus = torch.cat(mus, dim=1)
        log_pis = torch.cat(log_pis, dim=1)
        log_sigmas = torch.cat(log_sigmas, dim=1)
        # (B, T, num_gaussians)
        lf0_residuals = torch.cat(lf0_residuals, dim=1)

        if is_inference:
            mu = torch.cat(mus_inf, dim=1)
            # TODO: may need to track sigma. For now we only use mu
            return mu, mu
        else:
            return (log_pis, log_sigmas, mus), lf0_residuals

    def inference(self, x, lengths=None):
        return pad_inference(
            model=self,
            x=x,
            lengths=lengths,
            reduction_factor=self.reduction_factor,
            mdn=True,
        )


class BiLSTMResF0NonAttentiveDecoder(BaseModel):
    """BiLSTM-based encoder + duration-informed Tacotron with residual F0 prediction.

    Args:
        in_dim (int) : dimension of encoder hidden layer
        ff_hidden_dim (int): Hidden dimension of feed-forward layers in the encoder.
        conv_hidden_dim (int): Hidden dimension of convolution layers in the encoder.
        lstm_hidden_dim (int): Hidden dimension of LSTM layers in the encoder.
        num_lstm_layers (int): Number of LSTM layers in the encoder.
        out_dim (int) : dimension of output
        layers (int) : number of LSTM layers
        hidden_dim (int) : dimension of hidden layer
        prenet_layers (int) : number of pre-net layers
        prenet_hidden_dim (int) : dimension of pre-net hidden layer
        prenet_dropout (float) : dropout rate of pre-net
        zoneout (float) : zoneout rate
        reduction_factor (int) : reduction factor
        downsample_by_conv (bool) : if True, downsample by convolution
        scaled_tanh (bool) : if True, use scaled tanh for residual F0 prediction
        in_lf0_idx (int): index of lf0 in input features
        in_lf0_min (float): minimum of lf0 in the training data of input features
        in_lf0_max (float): maximum of lf0 in the training data of input features
        out_lf0_idx (int): index of lf0 in output features
        out_lf0_mean (float): mean of lf0 in the training data of output features
        out_lf0_scale (float): scale of lf0 in the training data of output features
        use_mdn (bool): if True, use mixture density network for F0 prediction
        num_gaussians (int): number of gaussians in MDN
        sampling_mode (str): sampling mode in inference. "mean" or "random"
        in_ph_start_idx (int): Start index of phoneme features.
        in_ph_end_idx (int): End index of phoneme features.
        embed_dim (int): Embedding dimension.
        init_type (str): initialization type
    """

    def __init__(
        self,
        in_dim=512,
        ff_hidden_dim=2048,
        conv_hidden_dim=1024,
        lstm_hidden_dim=256,
        num_lstm_layers=2,
        dropout=0.0,
        out_dim=80,
        decoder_layers=2,
        decoder_hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        scaled_tanh=True,
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
        use_mdn=False,
        num_gaussians=4,
        sampling_mode="mean",
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        init_type="none",
    ):
        super().__init__()
        self.reduction_factor = reduction_factor
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale
        self.use_mdn = use_mdn
        self.in_dim = in_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            in_ff_dim = embed_dim
        else:
            in_ff_dim = in_dim

        # Encoder
        # NOTE: can be simply replaced by a BiLSTM?
        # so far I use sinsy like architecture
        self.ff = nn.Sequential(
            nn.Linear(in_ff_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
        )

        self.conv = nn.Sequential(
            nn.ReflectionPad1d(3),
            nn.Conv1d(ff_hidden_dim + 1, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
        )

        self.lstm = nn.LSTM(
            conv_hidden_dim,
            lstm_hidden_dim,
            num_lstm_layers,
            bidirectional=True,
            batch_first=True,
            dropout=dropout,
        )

        # Autoregressive decoder
        decoder_in_dim = 2 * lstm_hidden_dim + 1
        # NOTE: hard code in_lf0_idx to -1
        cls = MDNResF0NonAttentiveDecoder if use_mdn else ResF0NonAttentiveDecoder
        if use_mdn:
            ex_kwargs = {"num_gaussians": num_gaussians, "sampling_mode": sampling_mode}
        else:
            ex_kwargs = {}
        self.decoder = cls(
            in_dim=decoder_in_dim,
            out_dim=out_dim,
            layers=decoder_layers,
            hidden_dim=decoder_hidden_dim,
            prenet_layers=prenet_layers,
            prenet_hidden_dim=prenet_hidden_dim,
            prenet_dropout=prenet_dropout,
            zoneout=zoneout,
            reduction_factor=reduction_factor,
            downsample_by_conv=downsample_by_conv,
            scaled_tanh=scaled_tanh,
            in_lf0_idx=-1,
            in_lf0_min=in_lf0_min,
            in_lf0_max=in_lf0_max,
            out_lf0_idx=out_lf0_idx,
            out_lf0_mean=out_lf0_mean,
            out_lf0_scale=out_lf0_scale,
            **ex_kwargs,
        )
        init_weights(self, init_type)

    def _set_lf0_params(self):
        # Special care for residual F0 prediction models
        # NOTE: don't overwrite out_lf0_idx and in_lf0_idx
        if hasattr(self.decoder, "out_lf0_mean"):
            self.decoder.in_lf0_min = self.in_lf0_min
            self.decoder.in_lf0_max = self.in_lf0_max
            self.decoder.out_lf0_mean = self.out_lf0_mean
            self.decoder.out_lf0_scale = self.out_lf0_scale

    def is_autoregressive(self):
        return self.decoder.is_autoregressive()

    def prediction_type(self):
        return (
            PredictionType.PROBABILISTIC
            if self.use_mdn
            else PredictionType.DETERMINISTIC
        )

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, x, lengths=None, y=None):
        self._set_lf0_params()
        lf0_score = x[:, :, self.in_lf0_idx].unsqueeze(-1)

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")

        out = self.ff(x)
        out = torch.cat([out, lf0_score], dim=-1)

        out = self.conv(out.transpose(1, 2)).transpose(1, 2)
        sequence = pack_padded_sequence(out, lengths, batch_first=True)
        out, _ = self.lstm(sequence)
        out, _ = pad_packed_sequence(out, batch_first=True)

        # NOTE: need to concat the lf0 score to the output of the lstm to tell
        # the decoder the lf0
        out = torch.cat([out, lf0_score], dim=-1)

        outs, lf0_residual = self.decoder(out, lengths, y)

        return outs, lf0_residual

    def inference(self, x, lengths=None):
        return pad_inference(
            model=self,
            x=x,
            lengths=lengths,
            reduction_factor=self.reduction_factor,
            mdn=self.use_mdn,
        )



from warnings import warn

import librosa
import numpy as np
import pyloudnorm as pyln
import pysptk
import pyworld
import scipy
import torch
from nnmnkwii.frontend import merlin as fe
from nnmnkwii.io import hts
from nnmnkwii.postfilters import merlin_post_filter
from nnmnkwii.preprocessing.f0 import interp1d
from nnsvs.base import PredictionType
from nnsvs.dsp import bandpass_filter
from nnsvs.io.hts import (
    get_note_frame_indices,
    get_note_indices,
    get_pitch_index,
    get_pitch_indices,
)
from nnsvs.multistream import (
    get_static_stream_sizes,
    get_windows,
    multi_stream_mlpg,
    split_streams,
)
from nnsvs.pitch import gen_sine_vibrato, lowpass_filter
from nnsvs.postfilters import variance_scaling
from sklearn.preprocessing import MinMaxScaler


def _midi_to_hz(x, idx, log_f0=False):
    z = np.zeros(len(x))
    indices = x[:, idx] > 0
    z[indices] = librosa.midi_to_hz(x[indices, idx])
    if log_f0:
        z[indices] = np.log(z[indices])
    return z


def _is_silence(label):
    is_full_context = "@" in label
    if is_full_context:
        is_silence = "-sil" in label or "-pau" in label
    else:
        is_silence = label == "sil" or label == "pau"
    return is_silence


@torch.no_grad()
def predict_timelag(
    device,
    labels,
    timelag_model,
    timelag_config,
    timelag_in_scaler,
    timelag_out_scaler,
    binary_dict,
    numeric_dict,
    pitch_indices=None,
    log_f0_conditioning=True,
    allowed_range=None,
    allowed_range_rest=None,
    force_clip_input_features=False,
    frame_period=5,
):
    """Predict time-lag from HTS labels

    Args:
        device (torch.device): device
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS-style labels
        timelag_model (nn.Module): time-lag model
        timelag_config (dict): time-lag model config
        timelag_in_scaler (sklearn.preprocessing.MinMaxScaler): input scaler
        timelag_out_scaler (sklearn.preprocessing.MinMaxScaler): output scaler
        binary_dict (dict): binary feature dict
        numeric_dict (dict): numeric feature dict
        pitch_indices (list): indices of pitch features
        log_f0_conditioning (bool): whether to condition on log f0
        allowed_range (list): allowed range of time-lag
        allowed_range_rest (list): allowed range of time-lag for rest
        force_clip_input_features (bool): whether to clip input features

    Returns;
        ndarray: time-lag predictions
    """
    hts_frame_shift = int(frame_period * 1e4)
    # make sure to set frame shift properly before calling round_ method
    labels.frame_shift = hts_frame_shift
    if pitch_indices is None:
        pitch_indices = get_pitch_indices(binary_dict, numeric_dict)

    if allowed_range is None:
        allowed_range = [-20, 20]
    if allowed_range_rest is None:
        allowed_range_rest = [-40, 40]
    # round start/end times just in case.
    labels.round_()

    # Extract note-level labels
    note_indices = get_note_indices(labels)
    note_labels = labels[note_indices]

    # Extract musical/linguistic context
    timelag_linguistic_features = fe.linguistic_features(
        note_labels,
        binary_dict,
        numeric_dict,
        add_frame_features=False,
        subphone_features=None,
        frame_shift=hts_frame_shift,
    ).astype(np.float32)

    # Adjust input features if we use log-f0 conditioning
    if log_f0_conditioning:
        if pitch_indices is None:
            raise ValueError("Pitch feature indices must be specified!")
        for idx in pitch_indices:
            timelag_linguistic_features[:, idx] = interp1d(
                _midi_to_hz(timelag_linguistic_features, idx, log_f0_conditioning),
                kind="slinear",
            )

    # Normalization
    timelag_linguistic_features = timelag_in_scaler.transform(
        timelag_linguistic_features
    )
    if force_clip_input_features and isinstance(timelag_in_scaler, MinMaxScaler):
        # clip to feature range (except for pitch-related features)
        non_pitch_indices = [
            idx
            for idx in range(timelag_linguistic_features.shape[1])
            if idx not in pitch_indices
        ]
        timelag_linguistic_features[:, non_pitch_indices] = np.clip(
            timelag_linguistic_features[:, non_pitch_indices],
            timelag_in_scaler.feature_range[0],
            timelag_in_scaler.feature_range[1],
        )

    # Run model
    x = torch.from_numpy(timelag_linguistic_features).unsqueeze(0).to(device)

    # Run model
    if timelag_model.prediction_type() == PredictionType.PROBABILISTIC:
        # (B, T, D_out)
        max_mu, max_sigma = timelag_model.inference(x, [x.shape[1]])
        if np.any(timelag_config.has_dynamic_features):
            # Apply denormalization
            # (B, T, D_out) -> (T, D_out)
            max_sigma_sq = (
                max_sigma.squeeze(0).cpu().data.numpy() ** 2 * timelag_out_scaler.var_
            )
            max_sigma_sq = np.maximum(max_sigma_sq, 1e-14)
            max_mu = timelag_out_scaler.inverse_transform(
                max_mu.squeeze(0).cpu().data.numpy()
            )
            # (T, D_out) -> (T, static_dim)
            pred_timelag = multi_stream_mlpg(
                max_mu,
                max_sigma_sq,
                get_windows(timelag_config.num_windows),
                timelag_config.stream_sizes,
                timelag_config.has_dynamic_features,
            )
        else:
            # Apply denormalization
            pred_timelag = timelag_out_scaler.inverse_transform(
                max_mu.squeeze(0).cpu().data.numpy()
            )
    else:
        # (T, D_out)
        pred_timelag = (
            timelag_model.inference(x, [x.shape[1]]).squeeze(0).cpu().data.numpy()
        )
        # Apply denormalization
        pred_timelag = timelag_out_scaler.inverse_transform(pred_timelag)
        if np.any(timelag_config.has_dynamic_features):
            # (T, D_out) -> (T, static_dim)
            pred_timelag = multi_stream_mlpg(
                pred_timelag,
                timelag_out_scaler.var_,
                get_windows(timelag_config.num_windows),
                timelag_config.stream_sizes,
                timelag_config.has_dynamic_features,
            )

    # Rounding
    pred_timelag = np.round(pred_timelag)

    # Clip to the allowed range
    for idx in range(len(pred_timelag)):
        if _is_silence(note_labels.contexts[idx]):
            pred_timelag[idx] = np.clip(
                pred_timelag[idx], allowed_range_rest[0], allowed_range_rest[1]
            )
        else:
            pred_timelag[idx] = np.clip(
                pred_timelag[idx], allowed_range[0], allowed_range[1]
            )

    # frames -> 100 ns
    pred_timelag *= hts_frame_shift

    return pred_timelag


@torch.no_grad()
def predict_duration(
    device,
    labels,
    duration_model,
    duration_config,
    duration_in_scaler,
    duration_out_scaler,
    binary_dict,
    numeric_dict,
    pitch_indices=None,
    log_f0_conditioning=True,
    force_clip_input_features=False,
    frame_period=5,
):
    """Predict phoneme durations from HTS labels

    Args:
        device (torch.device): device to run the model on
        labels (nnmnkwii.io.hts.HTSLabelFile): labels
        duration_model (nn.Module): duration model
        duration_config (dict): duration config
        duration_in_scaler (sklearn.preprocessing.MinMaxScaler): duration input scaler
        duration_out_scaler (sklearn.preprocessing.MinMaxScaler): duration output scaler
        binary_dict (dict): binary feature dictionary
        numeric_dict (dict): numeric feature dictionary
        pitch_indices (list): indices of pitch features
        log_f0_conditioning (bool): whether to use log-f0 conditioning
        force_clip_input_features (bool): whether to clip input features

    Returns:
        np.ndarray: predicted durations
    """
    hts_frame_shift = int(frame_period * 1e4)
    if pitch_indices is None:
        pitch_indices = get_pitch_indices(binary_dict, numeric_dict)

    # Extract musical/linguistic features
    duration_linguistic_features = fe.linguistic_features(
        labels,
        binary_dict,
        numeric_dict,
        add_frame_features=False,
        subphone_features=None,
        frame_shift=hts_frame_shift,
    ).astype(np.float32)

    if log_f0_conditioning:
        for idx in pitch_indices:
            duration_linguistic_features[:, idx] = interp1d(
                _midi_to_hz(duration_linguistic_features, idx, log_f0_conditioning),
                kind="slinear",
            )

    # Apply normalization
    duration_linguistic_features = duration_in_scaler.transform(
        duration_linguistic_features
    )
    if force_clip_input_features and isinstance(duration_in_scaler, MinMaxScaler):
        # clip to feature range (except for pitch-related features)
        non_pitch_indices = [
            idx
            for idx in range(duration_linguistic_features.shape[1])
            if idx not in pitch_indices
        ]
        duration_linguistic_features[:, non_pitch_indices] = np.clip(
            duration_linguistic_features[:, non_pitch_indices],
            duration_in_scaler.feature_range[0],
            duration_in_scaler.feature_range[1],
        )

    # Apply model
    x = torch.from_numpy(duration_linguistic_features).float().to(device)
    x = x.view(1, -1, x.size(-1))

    if duration_model.prediction_type() == PredictionType.PROBABILISTIC:
        # (B, T, D_out)
        max_mu, max_sigma = duration_model.inference(x, [x.shape[1]])
        if np.any(duration_config.has_dynamic_features):
            raise RuntimeError(
                "Dynamic features are not supported for duration modeling"
            )
        # Apply denormalization
        max_sigma_sq = (
            max_sigma.squeeze(0).cpu().data.numpy() ** 2 * duration_out_scaler.var_
        )
        max_sigma_sq = np.maximum(max_sigma_sq, 1e-14)
        max_mu = duration_out_scaler.inverse_transform(
            max_mu.squeeze(0).cpu().data.numpy()
        )

        return max_mu, max_sigma_sq
    else:
        # (T, D_out)
        pred_durations = (
            duration_model.inference(x, [x.shape[1]]).squeeze(0).cpu().data.numpy()
        )
        # Apply denormalization
        pred_durations = duration_out_scaler.inverse_transform(pred_durations)
        if np.any(duration_config.has_dynamic_features):
            # (T, D_out) -> (T, static_dim)
            pred_durations = multi_stream_mlpg(
                pred_durations,
                duration_out_scaler.var_,
                get_windows(duration_config.num_windows),
                duration_config.stream_sizes,
                duration_config.has_dynamic_features,
            )

    pred_durations[pred_durations <= 0] = 1
    pred_durations = np.round(pred_durations)

    return pred_durations


def postprocess_duration(labels, pred_durations, lag, frame_period=5):
    """Post-process durations based on predicted time-lag

    Ref : https://arxiv.org/abs/2108.02776

    Args:
        labels (HTSLabelFile): HTS labels
        pred_durations (array or tuple): predicted durations for non-MDN,
            mean and variance for MDN
        lag (array): predicted time-lag

    Returns:
        HTSLabelFile: labels with adjusted durations
    """
    hts_frame_shift = int(frame_period * 1e4)

    note_indices = get_note_indices(labels)
    # append the end of note
    note_indices.append(len(labels))

    is_mdn = isinstance(pred_durations, tuple) and len(pred_durations) == 2

    output_labels = hts.HTSLabelFile()

    for i in range(1, len(note_indices)):
        p = labels[note_indices[i - 1] : note_indices[i]]

        # Compute note duration with time-lag
        # eq (11)
        L = int(fe.duration_features(p)[0])
        if i < len(note_indices) - 1:
            L_hat = L - (lag[i - 1] - lag[i]) / hts_frame_shift
        else:
            L_hat = L - (lag[i - 1]) / hts_frame_shift

        # Prevent negative duration
        L_hat = max(L_hat, 1)

        # adjust the start time of the note
        p.start_times = np.minimum(
            np.asarray(p.start_times) + lag[i - 1].reshape(-1),
            np.asarray(p.end_times) - hts_frame_shift * len(p),
        )
        p.start_times = np.maximum(p.start_times, 0)
        if len(output_labels) > 0:
            p.start_times = np.maximum(
                p.start_times, output_labels.start_times[-1] + hts_frame_shift
            )

        # Compute normalized phoneme durations
        if is_mdn:
            mu = pred_durations[0][note_indices[i - 1] : note_indices[i]]
            sigma_sq = pred_durations[1][note_indices[i - 1] : note_indices[i]]
            # eq (17)
            rho = (L_hat - mu.sum()) / sigma_sq.sum()
            # eq (16)
            d_norm = mu + rho * sigma_sq

            if np.any(d_norm <= 0):
                # eq (12) (using mu as d_hat)
                s = frame_period * 0.001
                print(
                    f"Negative phoneme durations are predicted at {i}-th note. "
                    "The note duration: ",
                    f"{round(float(L)*s,3)} sec -> {round(float(L_hat)*s,3)} sec",
                )
                print(
                    "It's likely that the model couldn't predict correct durations "
                    "for short notes."
                )
                print(
                    f"Variance scaling based durations (in frame):\n{(mu + rho * sigma_sq)}"
                )
                print(
                    f"Fallback to uniform scaling (in frame):\n{(L_hat * mu / mu.sum())}"
                )
                d_norm = L_hat * mu / mu.sum()
        else:
            # eq (12)
            d_hat = pred_durations[note_indices[i - 1] : note_indices[i]]
            d_norm = L_hat * d_hat / d_hat.sum()

        d_norm = np.round(d_norm)
        d_norm[d_norm <= 0] = 1

        p.set_durations(d_norm)

        if len(output_labels) > 0:
            output_labels.end_times[-1] = p.start_times[0]
        for n in p:
            output_labels.append(n)

    return output_labels


@torch.no_grad()
def predict_timing(
    device,
    labels,
    binary_dict,
    numeric_dict,
    timelag_model,
    timelag_config,
    timelag_in_scaler,
    timelag_out_scaler,
    duration_model,
    duration_config,
    duration_in_scaler,
    duration_out_scaler,
    log_f0_conditioning=True,
    allowed_range=None,
    allowed_range_rest=None,
    force_clip_input_features=True,
    frame_period=5,
):
    """Predict timinigs from HTS labels

    This is equivalent to ``predict_timelag + predict_duration + postprocess_duration``.

    Args:
        device (torch.device): device to run the model on
        labels (nnmnkwii.io.hts.HTSLabelFile): labels
        binary_dict (dict): binary feature dictionary
        numeric_dict (dict): numeric feature dictionary
        timelag_model (nn.Module): timelag model
        timelag_config (dict): timelag config
        timelag_in_scaler (sklearn.preprocessing.MinMaxScaler): timelag input scaler
        timelag_out_scaler (sklearn.preprocessing.MinMaxScaler): timelag output scaler
        duration_model (nn.Module): duration model
        duration_config (dict): duration config
        duration_in_scaler (sklearn.preprocessing.MinMaxScaler): duration input scaler
        duration_out_scaler (sklearn.preprocessing.MinMaxScaler): duration output scaler
        log_f0_conditioning (bool): whether to condition on log f0
        allowed_range (list): allowed range of time-lag
        allowed_range_rest (list): allowed range of time-lag for rest
        force_clip_input_features (bool): whether to clip input features
        frame_period (int): frame period in milliseconds

    Returns:
        nnmnkwii.io.hts.HTSLabelFile: duration modified labels
    """
    hts_frame_shift = int(frame_period * 1e4)
    labels.frame_shift = hts_frame_shift

    pitch_indices = get_pitch_indices(binary_dict, numeric_dict)

    # Time-lag
    lag = predict_timelag(
        device=device,
        labels=labels,
        timelag_model=timelag_model,
        timelag_config=timelag_config,
        timelag_in_scaler=timelag_in_scaler,
        timelag_out_scaler=timelag_out_scaler,
        binary_dict=binary_dict,
        numeric_dict=numeric_dict,
        pitch_indices=pitch_indices,
        log_f0_conditioning=log_f0_conditioning,
        allowed_range=allowed_range,
        allowed_range_rest=allowed_range_rest,
        force_clip_input_features=force_clip_input_features,
        frame_period=frame_period,
    )

    # Duration predictions
    durations = predict_duration(
        device=device,
        labels=labels,
        duration_model=duration_model,
        duration_config=duration_config,
        duration_in_scaler=duration_in_scaler,
        duration_out_scaler=duration_out_scaler,
        binary_dict=binary_dict,
        numeric_dict=numeric_dict,
        pitch_indices=pitch_indices,
        log_f0_conditioning=log_f0_conditioning,
        force_clip_input_features=force_clip_input_features,
        frame_period=frame_period,
    )

    # Normalize phoneme durations
    duration_modified_labels = postprocess_duration(labels, durations, lag)
    return duration_modified_labels


@torch.no_grad()
def predict_acoustic(
    device,
    labels,
    acoustic_model,
    acoustic_config,
    acoustic_in_scaler,
    acoustic_out_scaler,
    binary_dict,
    numeric_dict,
    subphone_features="coarse_coding",
    pitch_indices=None,
    log_f0_conditioning=True,
    force_clip_input_features=False,
    frame_period=5,
    f0_shift_in_cent=0,
):
    """Predict acoustic features from HTS labels

    MLPG is applied to the predicted features if the output features have
    dynamic features.

    Args:
        device (torch.device): device to use
        labels (HTSLabelFile): HTS labels
        acoustic_model (nn.Module): acoustic model
        acoustic_config (AcousticConfig): acoustic configuration
        acoustic_in_scaler (sklearn.preprocessing.StandardScaler): input scaler
        acoustic_out_scaler (sklearn.preprocessing.StandardScaler): output scaler
        binary_dict (dict): binary feature dictionary
        numeric_dict (dict): numeric feature dictionary
        subphone_features (str): subphone feature type
        pitch_indices (list): indices of pitch features
        log_f0_conditioning (bool): whether to use log f0 conditioning
        force_clip_input_features (bool): whether to force clip input features
        frame_period (float): frame period in msec
        f0_shift_in_cent (float): F0 shift in cent-scale before the inference

    Returns:
        ndarray: predicted acoustic features
    """
    hts_frame_shift = int(frame_period * 1e4)
    if pitch_indices is None:
        pitch_indices = get_pitch_indices(binary_dict, numeric_dict)

    # Musical/linguistic features
    linguistic_features = fe.linguistic_features(
        labels,
        binary_dict,
        numeric_dict,
        add_frame_features=True,
        subphone_features=subphone_features,
        frame_shift=hts_frame_shift,
    )

    if log_f0_conditioning:
        for idx in pitch_indices:
            linguistic_features[:, idx] = interp1d(
                _midi_to_hz(linguistic_features, idx, log_f0_conditioning),
                kind="slinear",
            )
            if f0_shift_in_cent != 0:
                lf0_offset = f0_shift_in_cent * np.log(2) / 1200
                linguistic_features[:, idx] += lf0_offset

    # Apply normalization
    linguistic_features = acoustic_in_scaler.transform(linguistic_features)
    if force_clip_input_features and isinstance(acoustic_in_scaler, MinMaxScaler):
        # clip to feature range (except for pitch-related features)
        non_pitch_indices = [
            idx
            for idx in range(linguistic_features.shape[1])
            if idx not in pitch_indices
        ]
        linguistic_features[:, non_pitch_indices] = np.clip(
            linguistic_features[:, non_pitch_indices],
            acoustic_in_scaler.feature_range[0],
            acoustic_in_scaler.feature_range[1],
        )

    # Predict acoustic features
    x = torch.from_numpy(linguistic_features).float().to(device)
    x = x.view(1, -1, x.size(-1))

    if acoustic_model.prediction_type() in [
        PredictionType.PROBABILISTIC,
        PredictionType.MULTISTREAM_HYBRID,
    ]:
        # (B, T, D_out)
        max_mu, max_sigma = acoustic_model.inference(x, [x.shape[1]])
        if np.any(acoustic_config.has_dynamic_features):
            # Apply denormalization
            # (B, T, D_out) -> (T, D_out)
            max_sigma_sq = (
                max_sigma.squeeze(0).cpu().data.numpy() ** 2 * acoustic_out_scaler.var_
            )
            max_sigma_sq = np.maximum(max_sigma_sq, 1e-14)
            max_mu = acoustic_out_scaler.inverse_transform(
                max_mu.squeeze(0).cpu().data.numpy()
            )

            # (T, D_out) -> (T, static_dim)
            pred_acoustic = multi_stream_mlpg(
                max_mu,
                max_sigma_sq,
                get_windows(acoustic_config.num_windows),
                acoustic_config.stream_sizes,
                acoustic_config.has_dynamic_features,
            )
        else:
            # Apply denormalization
            pred_acoustic = acoustic_out_scaler.inverse_transform(
                max_mu.squeeze(0).cpu().data.numpy()
            )
    else:
        # (T, D_out)
        pred_acoustic = (
            acoustic_model.inference(x, [x.shape[1]]).squeeze(0).cpu().data.numpy()
        )
        # Apply denormalization
        pred_acoustic = acoustic_out_scaler.inverse_transform(pred_acoustic)
        if np.any(acoustic_config.has_dynamic_features):
            # (T, D_out) -> (T, static_dim)
            pred_acoustic = multi_stream_mlpg(
                pred_acoustic,
                acoustic_out_scaler.var_,
                get_windows(acoustic_config.num_windows),
                acoustic_config.stream_sizes,
                acoustic_config.has_dynamic_features,
            )

    return pred_acoustic


@torch.no_grad()
def postprocess_acoustic(
    device,
    acoustic_features,
    duration_modified_labels,
    binary_dict,
    numeric_dict,
    acoustic_config,
    acoustic_out_static_scaler,
    postfilter_model=None,
    postfilter_config=None,
    postfilter_out_scaler=None,
    sample_rate=48000,
    frame_period=5,
    relative_f0=False,
    feature_type="world",
    post_filter_type="gv",
    trajectory_smoothing=True,
    trajectory_smoothing_cutoff=50,
    trajectory_smoothing_cutoff_f0=20,
    vuv_threshold=0.5,
    f0_shift_in_cent=0,
    vibrato_scale=1.0,
    force_fix_vuv=False,
    fill_silence_to_rest=False,
):
    """Post-process acoustic features

    The function converts acoustic features in single ndarray to tuple of
    multi-stream acoustic features.

    e.g., array -> (mgc, lf0, vuv, bap)

    Args:
        device (torch.device): Device.
        duration_modified_labels (nnmnkwii.io.hts.HTSLabelFile): HTS label file.
        binary_dict (dict): Dictionary of binary features.
        numeric_dict (dict): Dictionary of numeric features.
        acoustic_config (dict): Acoustic model configuration.
        acoustic_features (np.ndarray): Acoustic features.
        acoustic_out_static_scaler (sklearn.preprocessing.StandardScaler): Scaler
            for acoustic features.
        postfilter_model (nn.Module): Post-filter model.
        postfilter_config (dict): Post-filter model configuration.
        postfilter_out_scaler (sklearn.preprocessing.StandardScaler): Scaler for post-filter.
        sample_rate (int): Sampling rate.
        frame_period (float): Frame period in milliseconds.
        relative_f0 (bool): If True, use relative f0.
        feature_type (str): Feature type.
        post_filter_type (str): Post-filter type.
            One of ``gv``, ``merlin`` or ``nnsvs``. Recommended to use ``gv``
            for general purpose.
        trajectory_smoothing (bool): Whether to apply trajectory smoothing.
        trajectory_smoothing_cutoff (float): Cutoff frequency for trajectory smoothing
            of spectral features.
        trajectory_smoothing_cutoff_f0 (float): Cutoff frequency for trajectory smoothing of f0.
        vuv_threshold (float): V/UV threshold.
        f0_shift_in_cent (float): F0 shift in cents.
        vibrato_scale (float): Vibrato scale.
        force_fix_vuv (bool): If True, force to fix V/UV.
        fill_silence_to_rest (bool): Fill silence to rest frames.

    Returns:
        tuple: Post-processed acoustic features.
    """
    hts_frame_shift = int(frame_period * 1e4)
    pitch_idx = get_pitch_index(binary_dict, numeric_dict)

    static_stream_sizes = get_static_stream_sizes(
        acoustic_config.stream_sizes,
        acoustic_config.has_dynamic_features,
        acoustic_config.num_windows,
    )

    linguistic_features = fe.linguistic_features(
        duration_modified_labels,
        binary_dict,
        numeric_dict,
        add_frame_features=True,
        frame_shift=hts_frame_shift,
    )
    # GV post-filter
    if post_filter_type == "gv" or (
        post_filter_type == "nnsvs" and feature_type == "world"
    ):
        note_frame_indices = get_note_frame_indices(
            binary_dict, numeric_dict, linguistic_features
        )
        if feature_type == "world":
            offset = 2
        elif feature_type == "melf0":
            # NOTE: set offset so that post-filter does not affect F0
            mel_freq = librosa.mel_frequencies(
                n_mels=80, fmin=63, fmax=sample_rate // 2
            )
            # NOTE: the threshold could be tuned for better performance
            offset = np.argmax(mel_freq > 1200)

        # NOTE: apply the post-filter for note frames only
        mgc_end_dim = static_stream_sizes[0]
        acoustic_features[:, :mgc_end_dim] = variance_scaling(
            acoustic_out_static_scaler.var_.reshape(-1)[:mgc_end_dim],
            acoustic_features[:, :mgc_end_dim],
            offset=offset,
            note_frame_indices=note_frame_indices,
        )

    # Learned post-filter using nnsvs
    if post_filter_type == "nnsvs" and postfilter_model is not None:
        # (1) Raw spectrogram or (2) mgc
        rawsp_output = postfilter_config.stream_sizes[0] >= 128

        # If the post-filter output is raw spectrogrma, convert mgc to log spectrogram
        if rawsp_output:
            outs = split_streams(acoustic_features, static_stream_sizes)
            assert len(outs) == 4
            mgc, lf0, vuv, bap = outs
            fft_size = pyworld.get_cheaptrick_fft_size(sample_rate)
            sp = pyworld.decode_spectral_envelope(
                mgc.astype(np.float64), sample_rate, fft_size
            ).astype(np.float32)
            sp = np.log(sp)
            acoustic_features = np.concatenate([sp, lf0, vuv, bap], axis=-1)

        in_feats = torch.from_numpy(acoustic_features).float().unsqueeze(0)
        in_feats = postfilter_out_scaler.transform(in_feats).float().to(device)
        # Run inference
        out_feats = postfilter_model.inference(in_feats, [in_feats.shape[1]])
        acoustic_features = (
            postfilter_out_scaler.inverse_transform(out_feats.cpu()).squeeze(0).numpy()
        )

        # Convert log spectrogram to mgc
        # NOTE: mgc is used to reduce possible artifacts
        # Ref: https://bit.ly/3AHjstU
        if rawsp_output:
            sp, lf0, vuv, bap = split_streams(
                acoustic_features, postfilter_config.stream_sizes
            )
            sp = np.exp(sp)
            mgc = pyworld.code_spectral_envelope(
                sp.astype(np.float64), sample_rate, 60
            ).astype(np.float32)
            acoustic_features = np.concatenate([mgc, lf0, vuv, bap], axis=-1)

    # Generate WORLD parameters
    if feature_type == "world":
        mgc, lf0, vuv, bap = gen_spsvs_static_features(
            labels=duration_modified_labels,
            acoustic_features=acoustic_features,
            binary_dict=binary_dict,
            numeric_dict=numeric_dict,
            stream_sizes=acoustic_config.stream_sizes,
            has_dynamic_features=acoustic_config.has_dynamic_features,
            pitch_idx=pitch_idx,
            num_windows=acoustic_config.num_windows,
            frame_period=frame_period,
            relative_f0=relative_f0,
            vibrato_scale=vibrato_scale,
            vuv_threshold=vuv_threshold,
            force_fix_vuv=force_fix_vuv,
        )
    elif feature_type == "melf0":
        mel, lf0, vuv = split_streams(acoustic_features, [80, 1, 1])
    else:
        raise ValueError(f"Unknown feature type: {feature_type}")

    if fill_silence_to_rest:
        mask = _get_nonrest_frame_soft_mask(
            binary_dict, numeric_dict, linguistic_features
        )
        if feature_type == "world":
            mgc, lf0, vuv, bap = _fill_silence_to_world_params(mgc, lf0, vuv, bap, mask)
        elif feature_type == "melf0":
            mel, lf0, vuv = _fill_silence_to_mel_params(mel, lf0, vuv, mask)

    if f0_shift_in_cent != 0:
        lf0_offset = f0_shift_in_cent * np.log(2) / 1200
        lf0 = lf0 + lf0_offset

    # NOTE: spectral enhancement based on the Merlin's post-filter implementation
    if feature_type == "world" and post_filter_type == "merlin":
        alpha = pysptk.util.mcepalpha(sample_rate)
        mgc = merlin_post_filter(mgc, alpha)

    # Remove high-frequency components of lf0/mgc/bap
    # NOTE: Useful to reduce high-frequency artifacts
    if trajectory_smoothing:
        modfs = int(1 / (frame_period * 0.001))
        lf0[:, 0] = lowpass_filter(
            lf0[:, 0], modfs, cutoff=trajectory_smoothing_cutoff_f0
        )
        if feature_type == "world":
            for d in range(mgc.shape[1]):
                mgc[:, d] = lowpass_filter(
                    mgc[:, d], modfs, cutoff=trajectory_smoothing_cutoff
                )
            for d in range(bap.shape[1]):
                bap[:, d] = lowpass_filter(
                    bap[:, d], modfs, cutoff=trajectory_smoothing_cutoff
                )
        elif feature_type == "melf0":
            for d in range(mel.shape[1]):
                mel[:, d] = lowpass_filter(
                    mel[:, d], modfs, cutoff=trajectory_smoothing_cutoff
                )
    if feature_type == "world":
        use_mcep_aperiodicity = bap.shape[-1] > 5
    if feature_type == "world" and not use_mcep_aperiodicity:
        bap = np.clip(bap, a_min=-60, a_max=0)

    if feature_type == "world":
        return mgc, lf0, vuv, bap
    elif feature_type == "melf0":
        return mel, lf0, vuv


@torch.no_grad()
def predict_waveform(
    device,
    multistream_features,
    vocoder=None,
    vocoder_config=None,
    vocoder_in_scaler=None,
    sample_rate=48000,
    frame_period=5,
    use_world_codec=True,
    feature_type="world",
    vocoder_type="world",
    vuv_threshold=0.5,
):
    """Predict waveform from multi-stream acoustic features

    Vocoders can be 1) WORLD, 2) PWG or 3) uSFGAN.

    Args:
        device (torch.device): Device to run inference
        features (tuple): Acoustic features
        vocoder (nn.Module): Vocoder model
        vocoder_config (dict): Vocoder config
        vocoder_in_scaler (StandardScaler): Vocoder input scaler
        sample_rate (int,): Sampling rate.
        frame_period (float): Frame period in msec.
        use_world_codec (bool): Whether to use WORLD codec for decoding.
        feature_type (str): Feature type.
            ``world`` ``world_org``, ``melf0`` or ``neutrino``.
        vocoder_type (str): Vocoder type. ``world`` or ``pwg`` or ``usfgan``
        vuv_threshold (float): VUV threshold.

    Returns:
        np.ndarray: Predicted waveform
    """
    if feature_type == "world":
        mgc, lf0, vuv, bap = multistream_features
    elif feature_type == "world_org":
        f0, spectrogram, aperiodicity = multistream_features
    elif feature_type == "neutrino":
        mgc, f0, bap = multistream_features
        # prepare (mgc, lf0, vuv, bap) to be compatible with NNSVS
        lf0 = f0.copy()
        lf0[np.nonzero(f0)] = np.log(f0[np.nonzero(f0)])
        vuv = (f0 > 0).astype(np.float32)
    elif feature_type == "melf0":
        mel, lf0, vuv = multistream_features
    else:
        raise ValueError(f"Unknown feature type: {feature_type}")

    # NOTE: `use_mcep_aperiodicity` was used for experimental purpose but didn't make
    # significant difference. Please just ignore or ping @r9y9 for details.
    if feature_type in ["world", "neutrino"]:
        use_mcep_aperiodicity = bap.shape[-1] > 5

    if feature_type == "neutrino" and not use_world_codec:
        raise ValueError("use_world_codec must be True when feature_type is neutrino")

    # Waveform generation by WORLD
    if vocoder_type == "world":
        if feature_type not in ["world", "world_org", "neutrino"]:
            raise ValueError(f"Invalid feature type for WORLD vocoder: {feature_type}")
        if feature_type == "world_org":
            # NOTE: WORLD-based features are already converted to raw WORLD parameters
            pass
        else:
            f0, spectrogram, aperiodicity = gen_world_params(
                mgc,
                lf0,
                vuv,
                bap,
                sample_rate,
                vuv_threshold=vuv_threshold,
                use_world_codec=use_world_codec,
            )
        # make sure to have float64 typed parameters
        wav = pyworld.synthesize(
            f0.astype(np.float64),
            spectrogram.astype(np.float64),
            aperiodicity.astype(np.float64),
            sample_rate,
            frame_period,
        )
    elif vocoder_type == "pwg":
        # NOTE: So far vocoder models are trained on binary V/UV features
        vuv = (vuv > vuv_threshold).astype(np.float32)
        if feature_type == "world":
            voc_inp = (
                torch.from_numpy(
                    vocoder_in_scaler.transform(
                        np.concatenate([mgc, lf0, vuv, bap], axis=-1)
                    )
                )
                .float()
                .to(device)
            )
        elif feature_type == "melf0":
            voc_inp = (
                torch.from_numpy(
                    vocoder_in_scaler.transform(
                        np.concatenate([mel, lf0, vuv], axis=-1)
                    )
                )
                .float()
                .to(device)
            )
        wav = vocoder.inference(voc_inp).view(-1).to("cpu").numpy()
    elif vocoder_type == "usfgan":
        if feature_type in ["world", "neutrino"]:
            fftlen = pyworld.get_cheaptrick_fft_size(sample_rate)
            if use_mcep_aperiodicity:
                # Convert mel-cepstrum-based aperiodicity to WORLD's aperiodicity
                aperiodicity_order = bap.shape[-1] - 1
                alpha = pysptk.util.mcepalpha(sample_rate)
                aperiodicity = pysptk.mc2sp(
                    np.ascontiguousarray(bap).astype(np.float64),
                    fftlen=fftlen,
                    alpha=alpha,
                )
            else:
                aperiodicity = pyworld.decode_aperiodicity(
                    np.ascontiguousarray(bap).astype(np.float64),
                    sample_rate,
                    fftlen,
                )
            # fill aperiodicity with ones for unvoiced regions
            aperiodicity[vuv.reshape(-1) < vuv_threshold, 0] = 1.0
            # WORLD fails catastrophically for out of range aperiodicity
            aperiodicity = np.clip(aperiodicity, 0.0, 1.0)

            # Convert aperiodicity back to BAP
            if use_mcep_aperiodicity:
                bap = pysptk.sp2mc(
                    aperiodicity,
                    order=aperiodicity_order,
                    alpha=alpha,
                )
            else:
                bap = pyworld.code_aperiodicity(aperiodicity, sample_rate).astype(
                    np.float32
                )
            aux_feats = [mgc, bap]
        elif feature_type == "melf0":
            aux_feats = [mel]
        elif feature_type == "world_org":
            # it is possible to implement here but I suppose nobody wants to use
            raise NotImplementedError()

        aux_feats = (
            torch.from_numpy(
                vocoder_in_scaler.transform(np.concatenate(aux_feats, axis=-1))
            )
            .float()
            .to(device)
        )

        contf0 = np.exp(lf0)
        if vocoder_config.data.sine_f0_type in ["contf0", "cf0"]:
            f0_inp = contf0
        elif vocoder_config.data.sine_f0_type == "f0":
            f0_inp = contf0
            f0_inp[vuv < vuv_threshold] = 0
        # NOTE: uSFGAN internally performs normalization
        # so we don't need to normalize inputs here
        wav = vocoder.inference(f0_inp, aux_feats).view(-1).to("cpu").numpy()

    return wav


def postprocess_waveform(
    wav,
    sample_rate,
    dtype=np.int16,
    peak_norm=False,
    loudness_norm=False,
    target_loudness=-20.0,
):
    """Perform post-processing for synthesized waveform

    Args:
        wav (ndarray): The input waveform
        sample_rate (int): The sampling rate
        dtype (np.dtype): The dtype of output waveform. Default is np.int16.
        peak_norm (bool): Whether to perform peak normalization
        loudness_norm (bool): Whether to perform loudness normalization
        target_loudness (float): Target loudness in dB

    Returns:
        ndarray: The post-processed waveform
    """
    wav = bandpass_filter(wav, sample_rate)

    # Peak normalize audio to 0 dB
    if peak_norm:
        wav = pyln.normalize.peak(wav, 0.0)

    # Normalize loudnes
    # NOTE: -20 dB is roughly the same as the NEURINO (NSF ver.)
    if loudness_norm:
        meter = pyln.Meter(sample_rate)
        loudness = meter.integrated_loudness(wav)
        wav = pyln.normalize.loudness(wav, loudness, target_loudness)

    # NOTE: use np.int16 to save disk space
    if dtype in [np.int16, "int16"]:
        # NOTE: NNSVS (>=0.0.3) uses waveforms normalized in [-1, 1] for training.
        # so the following code shouldn't be used but in case for models trained
        # with earlier NNSVS
        if np.max(np.abs(wav)) > 10:
            # data is likely already in [-32768, 32767]
            wav = wav.astype(np.int16)
        elif np.max(np.abs(wav)) <= 1:
            wav = (wav * 32767.0).astype(np.int16)
        else:
            # may need to handle int32 data (if any)
            warn("Unexpected waveform range: {} - {}".format(np.min(wav), np.max(wav)))
            warn("Failed to convert to int16. Returning waveform with floating point.")
    elif dtype is None:
        pass
    else:
        wav = wav.astype(dtype)

    return wav


def _get_nonrest_frame_soft_mask(
    binary_dict,
    numeric_dict,
    linguistic_features,
    win_length=200,
    duration_threshold=1.0,
):
    """Get mask for non-rest frames

    Args:
        binary_dict (dict): Dictionary for binary features
        numeric_dict (dict): Dictionary for numeric features
        linguistic_features (ndarray): Linguistic features
        win_length (int): Window length

    Returns:
        ndarray: Soft mask for non-rest frames.
            1 for non-rest frames and 0 for otherwise.
    """
    mask = np.ones(len(linguistic_features))

    in_sil_indices = []
    for k, v in binary_dict.items():
        name, _ = v
        if "C-Phone_sil" in name or "C-Phone_pau" in name:
            in_sil_indices.append(k)
    if len(in_sil_indices) == 0:
        return mask

    in_note_dur_idx = None
    for k, v in numeric_dict.items():
        name, _ = v
        if "e7" in name:
            in_note_dur_idx = k
            break

    dur = linguistic_features[:, len(binary_dict) + in_note_dur_idx]
    dur_in_sec = dur * 0.01
    for in_sil_idx in in_sil_indices:
        # Only mask out sil/pau segments over ${silence_threshold} sec. such as long pause
        mask[
            (linguistic_features[:, in_sil_idx] > 0) & (dur_in_sec > duration_threshold)
        ] = 0

    # make a smoothed mask with ${win_length} * 5ms window length
    mask = scipy.signal.convolve(mask, np.ones(win_length) / win_length, mode="same")

    # make sure that we don't mask out frames where notes are assigned
    pitch_idx = get_pitch_index(binary_dict, numeric_dict)
    score_f0 = linguistic_features[:, pitch_idx]
    mask[score_f0 > 0] = 1.0

    return mask.reshape(-1, 1)


def _fill_silence_to_world_params(mgc, lf0, vuv, bap, mask):
    mgc_sil = np.zeros((1, mgc.shape[1]))

    # NOTE: mgc_sil is a VERY ROUGH estimate of mgc for silence regions
    # the speech signal is assumed to be in [-1, 1].
    #   sr = 48000
    #   noise = np.random.randn(sr * 10) * 1e-5
    #   f0, timeaxis = pyworld.harvest(noise, sr, frame_period=5)
    #   f0[:] = 0
    #   spectrogram = pyworld.cheaptrick(noise, f0, timeaxis, sr)
    #   mgc = pyworld.code_spectral_envelope(spectrogram, sr, 60)
    #   print(mgc.mean(0))
    mgc_sil[0, 0] = -23.3
    mgc_sil[0, 1] = 0.0679
    mgc_sil[0, 2] = 0.00640
    mgc_sil[0, 3:] = 1e-3
    bap_sil = np.zeros_like(bap) + 1e-11

    mgc = mgc * mask + (1 - mask) * mgc_sil
    bap = bap * mask + (1 - mask) * bap_sil

    return mgc, lf0, vuv, bap


def _fill_silence_to_mel_params(mel, lf0, vuv, mask):
    # NOTE: -5.5 is also a very rough estimate of log-melspectrogram
    # for silence regions
    mel_sil = np.zeros((1, mel.shape[1])) - 5.5
    mel = mel * mask + (1 - mask) * mel_sil
    return mel, lf0, vuv


def correct_vuv_by_phone(vuv, binary_dict, linguistic_features):
    """Correct V/UV by phone-related flags in a hed file

    This function allows us to control V/UV explicitly by ``C-VUV_Voiced``
    and ``C-VUV_Unvoied`` flags in a hed file. This is useful when you see
    your trained acoustic model have lots of V/UV errors.
    Note that manually controlling V/UV means we are ignoring the
    acoustic model's prediction. It would have negative impact in some
    cases, but most cases it would help workaround V/UV errors.

    Args:
        vuv (ndarray): V/UV flags
        binary_dict (dict): binary feature dictionary
        linguistic_features (ndarray): linguistic features

    Returns:
        ndarray: corrected V/UV flags
    """
    vuv = vuv.copy()

    # Set V/UV to 1 based on the C-VUV_Voiced flag
    in_voiced_idx = -1
    for k, v in binary_dict.items():
        name, _ = v
        if "C-VUV_Voiced" in name:
            in_voiced_idx = k
            break
    if in_voiced_idx > 0:
        indices = linguistic_features[:, in_voiced_idx : in_voiced_idx + 1] > 0
        vuv[indices] = 1.0

    # Set V/UV to 0 based on the C-VUV_Unvoiced flag
    in_unvoiced_indices = []
    for k, v in binary_dict.items():
        name, _ = v
        if "C-VUV_Unvoiced" in name:
            in_unvoiced_indices.append(k)
    if len(in_unvoiced_indices) > 0:
        for in_unvoiced_idx in in_unvoiced_indices:
            indices = linguistic_features[:, in_unvoiced_idx : in_unvoiced_idx + 1] > 0
            vuv[indices] = 0.0

    # Set V/UV to 0 for sil/pau/br
    in_sil_indices = []
    for k, v in binary_dict.items():
        name, _ = v
        if "C-Phone_sil" in name or "C-Phone_pau" in name or "C-Phone_br" in name:
            in_sil_indices.append(k)
    if len(in_sil_indices) > 0:
        for in_sil_idx in in_sil_indices:
            indices = linguistic_features[:, in_sil_idx : in_sil_idx + 1] > 0
            vuv[indices] = 0.0

    return vuv


def gen_spsvs_static_features(
    labels,
    acoustic_features,
    binary_dict,
    numeric_dict,
    stream_sizes,
    has_dynamic_features,
    pitch_idx=None,
    num_windows=3,
    frame_period=5,
    relative_f0=True,
    vibrato_scale=1.0,
    vuv_threshold=0.3,
    force_fix_vuv=True,
):
    """Generate static features from predicted acoustic features

    Args:
        labels (HTSLabelFile): HTS labels
        acoustic_features (ndarray): predicted acoustic features
        binary_dict (dict): binary feature dictionary
        numeric_dict (dict): numeric feature dictionary
        stream_sizes (list): stream sizes
        has_dynamic_features (list): whether each stream has dynamic features
        pitch_idx (int): index of pitch features
        num_windows (int): number of windows
        frame_period (float): frame period
        relative_f0 (bool): whether to use relative f0
        vibrato_scale (float): vibrato scale
        vuv_threshold (float): vuv threshold
        force_fix_vuv (bool): whether to use post-processing to fix VUV.

    Returns:
        tuple: tuple of mgc, lf0, vuv and bap.
    """
    hts_frame_shift = int(frame_period * 1e4)
    if pitch_idx is None:
        pitch_idx = get_pitch_index(binary_dict, numeric_dict)

    if np.any(has_dynamic_features):
        static_stream_sizes = get_static_stream_sizes(
            stream_sizes, has_dynamic_features, num_windows
        )
    else:
        static_stream_sizes = stream_sizes

    # Copy here to avoid inplace operations on input acoustic features
    acoustic_features = acoustic_features.copy()

    # Split multi-stream features
    streams = split_streams(acoustic_features, static_stream_sizes)

    if len(streams) == 4:
        mgc, target_f0, vuv, bap = streams
        vib, vib_flags = None, None
    elif len(streams) == 5:
        # Assuming diff-based vibrato parameters
        mgc, target_f0, vuv, bap, vib = streams
        vib_flags = None
    elif len(streams) == 6:
        # Assuming sine-based vibrato parameters
        mgc, target_f0, vuv, bap, vib, vib_flags = streams
    else:
        raise RuntimeError("Not supported streams")

    linguistic_features = fe.linguistic_features(
        labels,
        binary_dict,
        numeric_dict,
        add_frame_features=True,
        frame_shift=hts_frame_shift,
    )

    # Correct V/UV based on special phone flags
    if force_fix_vuv:
        vuv = correct_vuv_by_phone(vuv, binary_dict, linguistic_features)

    # F0
    if relative_f0:
        diff_lf0 = target_f0
        f0_score = _midi_to_hz(linguistic_features, pitch_idx, False)[:, None]
        lf0_score = f0_score.copy()
        nonzero_indices = np.nonzero(lf0_score)
        lf0_score[nonzero_indices] = np.log(f0_score[nonzero_indices])
        lf0_score = interp1d(lf0_score, kind="slinear")

        f0 = diff_lf0 + lf0_score
        f0[vuv < vuv_threshold] = 0
        f0[np.nonzero(f0)] = np.exp(f0[np.nonzero(f0)])
    else:
        f0 = target_f0
        f0[vuv < vuv_threshold] = 0
        f0[np.nonzero(f0)] = np.exp(f0[np.nonzero(f0)])

    if vib is not None:
        if vib_flags is not None:
            # Generate sine-based vibrato
            vib_flags = vib_flags.flatten()
            m_a, m_f = vib[:, 0], vib[:, 1]

            # Fill zeros for non-vibrato frames
            m_a[vib_flags < 0.5] = 0
            m_f[vib_flags < 0.5] = 0

            # Gen vibrato
            sr_f0 = int(1 / (frame_period * 0.001))
            f0 = gen_sine_vibrato(f0.flatten(), sr_f0, m_a, m_f, vibrato_scale)
        else:
            # Generate diff-based vibrato
            f0 = f0.flatten() + vibrato_scale * vib.flatten()

    # NOTE: Back to log-domain for convenience
    lf0 = f0.copy()
    lf0[np.nonzero(lf0)] = np.log(f0[np.nonzero(lf0)])
    # NOTE: interpolation is necessary
    lf0 = interp1d(lf0, kind="slinear")

    lf0 = lf0[:, None] if len(lf0.shape) == 1 else lf0
    vuv = vuv[:, None] if len(vuv.shape) == 1 else vuv

    return mgc, lf0, vuv, bap


def gen_world_params(
    mgc,
    lf0,
    vuv,
    bap,
    sample_rate,
    vuv_threshold=0.3,
    use_world_codec=False,
):
    """Generate WORLD parameters from mgc, lf0, vuv and bap.

    Args:
        mgc (ndarray): mgc
        lf0 (ndarray): lf0
        vuv (ndarray): vuv
        bap (ndarray): bap
        sample_rate (int): sample rate
        vuv_threshold (float): threshold for VUV
        use_world_codec (bool): whether to use WORLD codec for spectral envelope

    Returns:
        tuple: tuple of f0, spectrogram and aperiodicity
    """
    fftlen = pyworld.get_cheaptrick_fft_size(sample_rate)
    alpha = pysptk.util.mcepalpha(sample_rate)
    use_mcep_aperiodicity = bap.shape[-1] > 5

    if use_world_codec:
        spectrogram = pyworld.decode_spectral_envelope(
            np.ascontiguousarray(mgc).astype(np.float64), sample_rate, fftlen
        )
    else:
        spectrogram = pysptk.mc2sp(
            np.ascontiguousarray(mgc), fftlen=fftlen, alpha=alpha
        )

    if use_mcep_aperiodicity:
        aperiodicity = pysptk.mc2sp(
            np.ascontiguousarray(bap), fftlen=fftlen, alpha=alpha
        )
    else:
        aperiodicity = pyworld.decode_aperiodicity(
            np.ascontiguousarray(bap).astype(np.float64), sample_rate, fftlen
        )

    # fill aperiodicity with ones for unvoiced regions
    aperiodicity[vuv.reshape(-1) < vuv_threshold, 0] = 1.0
    # WORLD fails catastrophically for out of range aperiodicity
    aperiodicity = np.clip(aperiodicity, 0.0, 1.0)

    f0 = lf0.copy()
    f0[np.nonzero(f0)] = np.exp(f0[np.nonzero(f0)])
    f0[vuv < vuv_threshold] = 0

    f0 = f0.flatten().astype(np.float64)
    spectrogram = spectrogram.astype(np.float64)
    aperiodicity = aperiodicity.astype(np.float64)

    return f0, spectrogram, aperiodicity



import os
import shutil
import tarfile
from os.path import join
from pathlib import Path
from urllib.request import urlretrieve

from nnsvs.util import dynamic_import
from tqdm.auto import tqdm

DEFAULT_CACHE_DIR = join(os.path.expanduser("~"), ".cache", "nnsvs")
CACHE_DIR = os.environ.get("NNSVS_CACHE_DIR", DEFAULT_CACHE_DIR)

model_registry = {
    "r9y9/yoko_latest": {
        "url": "https://www.dropbox.com/s/k8mya65yt52m0ps/yoko_latest.tar.gz?dl=1",
        "_target_": "nnsvs.svs:SPSVS",
    },
    "r9y9/20220322_yoko_timelag_mdn_duration_mdn_acoustic_resf0conv": {
        "url": "https://www.dropbox.com/s/olsfyqol9ryk5kx/"
        "20220322_yoko_timelag_mdn_duration_mdn_acoustic_resf0conv.tar.gz?dl=1",
        "_target_": "nnsvs.svs:SPSVS",
    },
}


def create_svs_engine(name, *args, **kwargs):
    """Create SVS engine from pretrained models.

    Args:
        name (str): Pre-trained model name
        args (list): Additional args for instantiation
        kwargs (dict): Additional kwargs for instantiation

    Returns:
        object: instance of SVS engine
    """
    if name not in model_registry:
        s = ""
        for model_id in get_available_model_ids():
            s += f"'{model_id}'\n"
        raise ValueError(
            f"""
Pretrained model '{name}' does not exist!

Available models:
{s[:-1]}"""
        )

    # download if not exists
    model_dir = retrieve_pretrained_model(name)

    # create an instance
    return dynamic_import(model_registry[name]["_target_"])(model_dir, *args, **kwargs)


def get_available_model_ids():
    """Get available pretrained model names.

    Returns:
        list: List of available pretrained model names.
    """
    return list(model_registry.keys())


# https://github.com/tqdm/tqdm#hooks-and-callbacks
class _TqdmUpTo(tqdm):  # type: ignore
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)


def is_pretrained_model_ready(name):
    out_dir = Path(CACHE_DIR) / name
    if out_dir.exists() and len(list(out_dir.glob("*.pth"))) == 0:
        return False
    return out_dir.exists()


def retrieve_pretrained_model(name):
    """Retrieve pretrained model from local cache or download from GitHub.

    Args:
        name (str): Name of pretrained model.

    Returns:
        str: Path to the pretrained model.

    Raises:
        ValueError: If the pretrained model is not found.
    """
    global model_registry
    if name not in model_registry:
        s = ""
        for model_id in get_available_model_ids():
            s += f"'{model_id}'\n"
        raise ValueError(
            f"""
Pretrained model '{name}' does not exist!

Available models:
{s[:-1]}"""
        )

    url = model_registry[name]["url"]
    # NOTE: assuming that filename and extracted is the same
    out_dir = Path(CACHE_DIR) / name
    model_dir = out_dir.parent
    out_dir.mkdir(parents=True, exist_ok=True)
    filename = Path(CACHE_DIR) / f"{name}.tar.gz"

    # re-download models
    if out_dir.exists() and len(list(out_dir.glob("*.pth"))) == 0:
        shutil.rmtree(out_dir)

    if not out_dir.exists():
        print('Downloading: "{}"'.format(url))
        with _TqdmUpTo(
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            miniters=1,
            desc=f"{name}.tar.gz",
        ) as t:  # all optional kwargs
            urlretrieve(url, filename, reporthook=t.update_to)
            t.total = t.n
        with tarfile.open(filename, mode="r|gz") as f:
            f.extractall(path=model_dir)
        os.remove(filename)

    return out_dir



# -*- coding: utf-8 -*-

# Copyright 2022 Reo Yoneyama (Nagoya University)
#  MIT License (https://opensource.org/licenses/MIT)

"""Residual block modules.

References:
    - https://github.com/bigpon/QPPWG
    - https://github.com/kan-bayashi/ParallelWaveGAN
    - https://github.com/r9y9/wavenet_vocoder

"""

import math
import sys
from logging import getLogger

import torch
import torch.nn as nn
from nnsvs.usfgan.utils import pd_indexing

# A logger for this file
logger = getLogger(__name__)


class Conv1d(nn.Conv1d):
    """Conv1d module with customized initialization."""

    def __init__(self, *args, **kwargs):
        """Initialize Conv1d module."""
        super(Conv1d, self).__init__(*args, **kwargs)

    def reset_parameters(self):
        """Reset parameters."""
        nn.init.kaiming_normal_(self.weight, nonlinearity="relu")
        if self.bias is not None:
            nn.init.constant_(self.bias, 0.0)


class Conv1d1x1(Conv1d):
    """1x1 Conv1d with customized initialization."""

    def __init__(self, in_channels, out_channels, bias=True):
        """Initialize 1x1 Conv1d module."""
        super(Conv1d1x1, self).__init__(
            in_channels, out_channels, kernel_size=1, padding=0, dilation=1, bias=bias
        )


class Conv2d(nn.Conv2d):
    """Conv2d module with customized initialization."""

    def __init__(self, *args, **kwargs):
        """Initialize Conv2d module."""
        super(Conv2d, self).__init__(*args, **kwargs)

    def reset_parameters(self):
        """Reset parameters."""
        nn.init.kaiming_normal_(self.weight, mode="fan_out", nonlinearity="relu")
        if self.bias is not None:
            nn.init.constant_(self.bias, 0.0)


class Conv2d1x1(Conv2d):
    """1x1 Conv2d with customized initialization."""

    def __init__(self, in_channels, out_channels, bias=True):
        """Initialize 1x1 Conv2d module."""
        super(Conv2d1x1, self).__init__(
            in_channels, out_channels, kernel_size=1, padding=0, dilation=1, bias=bias
        )


class FixedBlock(nn.Module):
    """Fixed block module in QPPWG."""

    def __init__(
        self,
        residual_channels=64,
        gate_channels=128,
        skip_channels=64,
        aux_channels=80,
        kernel_size=3,
        dilation=1,
        bias=True,
    ):
        """Initialize Fixed ResidualBlock module.

        Args:
            residual_channels (int): Number of channels for residual connection.
            skip_channels (int): Number of channels for skip connection.
            aux_channels (int): Local conditioning channels i.e. auxiliary input dimension.
            dilation (int): Dilation size.
            bias (bool): Whether to add bias parameter in convolution layers.

        """
        super(FixedBlock, self).__init__()
        padding = (kernel_size - 1) // 2 * dilation

        # dilation conv
        self.conv = Conv1d(
            residual_channels,
            gate_channels,
            kernel_size,
            padding=padding,
            padding_mode="reflect",
            dilation=dilation,
            bias=bias,
        )

        # local conditioning
        if aux_channels > 0:
            self.conv1x1_aux = Conv1d1x1(aux_channels, gate_channels, bias=False)
        else:
            self.conv1x1_aux = None

        # conv output is split into two groups
        gate_out_channels = gate_channels // 2
        self.conv1x1_out = Conv1d1x1(gate_out_channels, residual_channels, bias=bias)
        self.conv1x1_skip = Conv1d1x1(gate_out_channels, skip_channels, bias=bias)

    def forward(self, x, c):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input tensor (B, residual_channels, T).
            c (Tensor): Local conditioning auxiliary tensor (B, aux_channels, T).

        Returns:
            Tensor: Output tensor for residual connection (B, residual_channels, T).
            Tensor: Output tensor for skip connection (B, skip_channels, T).

        """
        residual = x
        x = self.conv(x)

        # split into two part for gated activation
        splitdim = 1
        xa, xb = x.split(x.size(splitdim) // 2, dim=splitdim)

        # local conditioning
        if c is not None:
            assert self.conv1x1_aux is not None
            c = self.conv1x1_aux(c)
            ca, cb = c.split(c.size(splitdim) // 2, dim=splitdim)
            xa, xb = xa + ca, xb + cb

        x = torch.tanh(xa) * torch.sigmoid(xb)

        # for skip connection
        s = self.conv1x1_skip(x)

        # for residual connection
        x = (self.conv1x1_out(x) + residual) * math.sqrt(0.5)

        return x, s


class AdaptiveBlock(nn.Module):
    """Adaptive block module in QPPWG."""

    def __init__(
        self,
        residual_channels=64,
        gate_channels=128,
        skip_channels=64,
        aux_channels=80,
        bias=True,
    ):
        """Initialize Adaptive ResidualBlock module.

        Args:
            residual_channels (int): Number of channels for residual connection.
            skip_channels (int): Number of channels for skip connection.
            aux_channels (int): Local conditioning channels i.e. auxiliary input dimension.
            bias (bool): Whether to add bias parameter in convolution layers.

        """
        super(AdaptiveBlock, self).__init__()

        # pitch-dependent dilation conv
        self.convP = Conv1d1x1(residual_channels, gate_channels, bias=bias)  # past
        self.convC = Conv1d1x1(residual_channels, gate_channels, bias=bias)  # current
        self.convF = Conv1d1x1(residual_channels, gate_channels, bias=bias)  # future

        # local conditioning
        if aux_channels > 0:
            self.conv1x1_aux = Conv1d1x1(aux_channels, gate_channels, bias=False)
        else:
            self.conv1x1_aux = None

        # conv output is split into two groups
        gate_out_channels = gate_channels // 2
        self.conv1x1_out = Conv1d1x1(gate_out_channels, residual_channels, bias=bias)
        self.conv1x1_skip = Conv1d1x1(gate_out_channels, skip_channels, bias=bias)

    def forward(self, xC, xP, xF, c):
        """Calculate forward propagation.

        Args:
            xC (Tensor): Current input tensor (B, residual_channels, T).
            xP (Tensor): Past input tensor (B, residual_channels, T).
            xF (Tensor): Future input tensor (B, residual_channels, T).
            c (Tensor): Local conditioning auxiliary tensor (B, aux_channels, T).

        Returns:
            Tensor: Output tensor for residual connection (B, residual_channels, T).
            Tensor: Output tensor for skip connection (B, skip_channels, T).

        """
        residual = xC
        x = self.convC(xC) + self.convP(xP) + self.convF(xF)

        # split into two part for gated activation
        splitdim = 1
        xa, xb = x.split(x.size(splitdim) // 2, dim=splitdim)

        # local conditioning
        if c is not None:
            assert self.conv1x1_aux is not None
            c = self.conv1x1_aux(c)
            ca, cb = c.split(c.size(splitdim) // 2, dim=splitdim)
            xa, xb = xa + ca, xb + cb

        x = torch.tanh(xa) * torch.sigmoid(xb)

        # for skip connection
        s = self.conv1x1_skip(x)

        # for residual connection
        x = (self.conv1x1_out(x) + residual) * math.sqrt(0.5)

        return x, s


class ResidualBlocks(nn.Module):
    """Multiple residual blocks stacking module."""

    def __init__(
        self,
        blockA,
        cycleA,
        blockF,
        cycleF,
        cascade_mode=0,
        residual_channels=64,
        gate_channels=128,
        skip_channels=64,
        aux_channels=80,
    ):
        """Initialize ResidualBlocks module.

        Args:
            blockA (int): Number of adaptive residual blocks.
            cycleA (int): Number of dilation cycles of adaptive residual blocks.
            blockF (int): Number of fixed residual blocks.
            cycleF (int): Number of dilation cycles of fixed residual blocks.
            cascade_mode (int): Cascaded mode (0: Adaptive->Fixed; 1: Fixed->Adaptive).
            residual_channels (int): Number of channels in residual conv.
            gate_channels (int):  Number of channels in gated conv.
            skip_channels (int): Number of channels in skip conv.
            aux_channels (int): Number of channels for auxiliary feature conv.

        """
        super(ResidualBlocks, self).__init__()

        # check the number of blocks and cycles
        cycleA = max(cycleA, 1)
        cycleF = max(cycleF, 1)
        assert blockA % cycleA == 0
        self.blockA_per_cycle = blockA // cycleA
        assert blockF % cycleF == 0
        blockF_per_cycle = blockF // cycleF

        # define adaptive residual blocks
        adaptive_blocks = nn.ModuleList()
        for block in range(blockA):
            conv = AdaptiveBlock(
                residual_channels=residual_channels,
                gate_channels=gate_channels,
                skip_channels=skip_channels,
                aux_channels=aux_channels,
            )
            adaptive_blocks += [conv]

        # define fixed residual blocks
        fixed_blocks = nn.ModuleList()
        for block in range(blockF):
            dilation = 2 ** (block % blockF_per_cycle)
            conv = FixedBlock(
                residual_channels=residual_channels,
                gate_channels=gate_channels,
                skip_channels=skip_channels,
                aux_channels=aux_channels,
                dilation=dilation,
            )
            fixed_blocks += [conv]

        # define cascaded structure
        if cascade_mode == 0:  # adaptive->fixed
            self.conv_dilated = adaptive_blocks.extend(fixed_blocks)
            self.block_modes = [True] * blockA + [False] * blockF
        elif cascade_mode == 1:  # fixed->adaptive
            self.conv_dilated = fixed_blocks.extend(adaptive_blocks)
            self.block_modes = [False] * blockF + [True] * blockA
        else:
            logger.error(f"Cascaded mode {cascade_mode} is not supported!")
            sys.exit(0)

    def forward(self, x, c, d, batch_index, ch_index):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            c (Tensor): Local conditioning auxiliary features (B, C ,T).
            d (Tensor): Input pitch-dependent dilated factors (B, 1, T).

        Returns:
            Tensor: Output tensor (B, residual_channels, T).

        """
        skips = 0
        blockA_idx = 0
        for f, mode in zip(self.conv_dilated, self.block_modes):
            if mode:  # adaptive block
                dilation = 2 ** (blockA_idx % self.blockA_per_cycle)
                xP, xF = pd_indexing(x, d, dilation, batch_index, ch_index)
                x, h = f(x, xP, xF, c)
                blockA_idx += 1
            else:  # fixed block
                x, h = f(x, c)
            skips = h + skips
        skips *= math.sqrt(1.0 / len(self.conv_dilated))

        return x


class PeriodicityEstimator(nn.Module):
    """Periodicity estimator module."""

    def __init__(
        self,
        in_channels,
        residual_channels=64,
        conv_layers=3,
        kernel_size=5,
        dilation=1,
        padding_mode="replicate",
    ):
        """Initialize USFGANGenerator module.

        Args:
            in_channels (int): Number of input channels.
            residual_channels (int): Number of channels in residual conv.
            conv_layers (int):  # Number of convolution layers.
            kernel_size (int): Kernel size.
            dilation (int): Dilation size.
            padding_mode (str): Padding mode.

        """
        super(PeriodicityEstimator, self).__init__()

        modules = []
        for idx in range(conv_layers):
            conv1d = Conv1d(
                in_channels,
                residual_channels,
                kernel_size=kernel_size,
                dilation=dilation,
                padding=kernel_size // 2 * dilation,
                padding_mode=padding_mode,
            )

            # initialize the initial outputs sigmoid(0)=0.5 to stabilize training
            if idx != conv_layers - 1:
                nonlinear = nn.ReLU(inplace=True)
            else:
                # NOTE: zero init induces nan or inf if weight normalization is used
                # nn.init.zeros_(conv1d.weight)
                nn.init.normal_(conv1d.weight, std=1e-4)
                nonlinear = nn.Sigmoid()

            modules += [conv1d, nonlinear]
            in_channels = residual_channels

        self.layers = nn.Sequential(*modules)

    def forward(self, x):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input auxiliary features (B, C ,T).

        Returns:
            Tensor: Output tensor (B, residual_channels, T).

        """
        return self.layers(x)



import os
from os.path import join

import hydra
import numpy as np
import pysptk
import pyworld
import torch
from hydra.utils import to_absolute_path
from nnsvs.dsp import bandpass_filter
from nnsvs.gen import gen_world_params
from nnsvs.logger import getLogger
from nnsvs.multistream import get_static_stream_sizes, split_streams
from nnsvs.svs import load_vocoder
from nnsvs.util import init_seed, load_utt_list
from omegaconf import DictConfig, OmegaConf
from scipy.io import wavfile
from tqdm.auto import tqdm


@torch.no_grad()
def anasyn(
    device,
    acoustic_features,
    acoustic_config,
    vocoder=None,
    vocoder_config=None,
    vocoder_in_scaler=None,
    sample_rate=48000,
    frame_period=5,
    vuv_threshold=0.5,
    use_world_codec=True,
    feature_type="world",
    vocoder_type="world",
):
    static_stream_sizes = get_static_stream_sizes(
        acoustic_config.stream_sizes,
        acoustic_config.has_dynamic_features,
        acoustic_config.num_windows,
    )

    # Split multi-stream features
    streams = split_streams(acoustic_features, static_stream_sizes)

    # Generate WORLD parameters
    if feature_type == "world":
        assert len(streams) == 4
        mgc, lf0, vuv, bap = streams
    elif feature_type == "melf0":
        mel, lf0, vuv = split_streams(acoustic_features, [80, 1, 1])
    else:
        raise ValueError(f"Unknown feature type: {feature_type}")

    # Waveform generation by (1) WORLD or (2) neural vocoder
    if vocoder_type == "world":
        f0, spectrogram, aperiodicity = gen_world_params(
            mgc,
            lf0,
            vuv,
            bap,
            sample_rate,
            vuv_threshold=vuv_threshold,
            use_world_codec=use_world_codec,
        )
        wav = pyworld.synthesize(
            f0,
            spectrogram,
            aperiodicity,
            sample_rate,
            frame_period,
        )
    elif vocoder_type == "pwg":
        # NOTE: So far vocoder models are trained on binary V/UV features
        vuv = (vuv > vuv_threshold).astype(np.float32)
        if feature_type == "world":
            voc_inp = (
                torch.from_numpy(
                    vocoder_in_scaler.transform(
                        np.concatenate([mgc, lf0, vuv, bap], axis=-1)
                    )
                )
                .float()
                .to(device)
            )
        elif feature_type == "melf0":
            voc_inp = (
                torch.from_numpy(
                    vocoder_in_scaler.transform(
                        np.concatenate([mel, lf0, vuv], axis=-1)
                    )
                )
                .float()
                .to(device)
            )
        wav = vocoder.inference(voc_inp).view(-1).to("cpu").numpy()
    elif vocoder_type == "usfgan":
        if feature_type == "world":
            fftlen = pyworld.get_cheaptrick_fft_size(sample_rate)
            use_mcep_aperiodicity = bap.shape[-1] > 5
            if use_mcep_aperiodicity:
                mcep_aperiodicity_order = bap.shape[-1] - 1
                alpha = pysptk.util.mcepalpha(sample_rate)
                aperiodicity = pysptk.mc2sp(
                    np.ascontiguousarray(bap).astype(np.float64),
                    fftlen=fftlen,
                    alpha=alpha,
                )
            else:
                aperiodicity = pyworld.decode_aperiodicity(
                    np.ascontiguousarray(bap).astype(np.float64), sample_rate, fftlen
                )
            # fill aperiodicity with ones for unvoiced regions
            aperiodicity[vuv.reshape(-1) < vuv_threshold, 0] = 1.0
            # WORLD fails catastrophically for out of range aperiodicity
            aperiodicity = np.clip(aperiodicity, 0.0, 1.0)
            # back to bap
            if use_mcep_aperiodicity:
                bap = pysptk.sp2mc(
                    aperiodicity,
                    order=mcep_aperiodicity_order,
                    alpha=alpha,
                )
            else:
                bap = pyworld.code_aperiodicity(aperiodicity, sample_rate).astype(
                    np.float32
                )

            aux_feats = (
                torch.from_numpy(
                    vocoder_in_scaler.transform(np.concatenate([mgc, bap], axis=-1))
                )
                .float()
                .to(device)
            )
        elif feature_type == "melf0":
            # NOTE: So far vocoder models are trained on binary V/UV features
            vuv = (vuv > vuv_threshold).astype(np.float32)
            aux_feats = (
                torch.from_numpy(vocoder_in_scaler.transform(mel)).float().to(device)
            )
        contf0 = np.exp(lf0)
        if vocoder_config.data.sine_f0_type in ["contf0", "cf0"]:
            f0_inp = contf0
        elif vocoder_config.data.sine_f0_type == "f0":
            f0_inp = contf0
            f0_inp[vuv < vuv_threshold] = 0
        wav = vocoder.inference(f0_inp, aux_feats).view(-1).to("cpu").numpy()

    return wav


def post_process(wav, sample_rate):
    wav = bandpass_filter(wav, sample_rate)

    if np.max(wav) > 10:
        if np.abs(wav).max() > 32767:
            wav = wav / np.abs(wav).max()
        # data is likely already in [-32768, 32767]
        wav = wav.astype(np.int16)
    else:
        if np.abs(wav).max() > 1.0:
            wav = wav / np.abs(wav).max()
        wav = (wav * 32767.0).astype(np.int16)
    return wav


@hydra.main(config_path="conf/synthesis", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    if not torch.cuda.is_available():
        device = torch.device("cpu")
    else:
        device = torch.device(config.device)

    acoustic_config = OmegaConf.load(to_absolute_path(config.acoustic.model_yaml))

    # Vocoder
    if config.vocoder.checkpoint is not None and len(config.vocoder.checkpoint) > 0:
        vocoder, vocoder_in_scaler, vocoder_config = load_vocoder(
            to_absolute_path(config.vocoder.checkpoint),
            device,
            acoustic_config,
        )
    else:
        vocoder, vocoder_in_scaler, vocoder_config = None, None, None
        if config.synthesis.vocoder_type != "world":
            logger.warning("Vocoder checkpoint is not specified")
            logger.info(f"Use world instead of {config.synthesis.vocoder_type}.")
        config.synthesis.vocoder_type = "world"

    # Run synthesis for each utt.

    in_dir = to_absolute_path(config.in_dir)
    out_dir = to_absolute_path(config.out_dir)
    os.makedirs(out_dir, exist_ok=True)
    utt_ids = load_utt_list(to_absolute_path(config.utt_list))
    logger.info("Processes %s utterances...", len(utt_ids))
    for utt_id in tqdm(utt_ids):
        acoustic_features = np.load(join(in_dir, f"{utt_id}-feats.npy"))
        init_seed(1234)

        wav = anasyn(
            device=device,
            acoustic_features=acoustic_features,
            acoustic_config=acoustic_config,
            vocoder=vocoder,
            vocoder_config=vocoder_config,
            vocoder_in_scaler=vocoder_in_scaler,
            sample_rate=config.synthesis.sample_rate,
            frame_period=config.synthesis.frame_period,
            use_world_codec=config.synthesis.use_world_codec,
            feature_type=config.synthesis.feature_type,
            vocoder_type=config.synthesis.vocoder_type,
            vuv_threshold=config.synthesis.vuv_threshold,
        )
        wav = post_process(wav, config.synthesis.sample_rate)
        out_wav_path = join(out_dir, f"{utt_id}.wav")
        wavfile.write(
            out_wav_path, rate=config.synthesis.sample_rate, data=wav.astype(np.int16)
        )


def entry():
    my_app()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-

# Copyright 2020 Yi-Chiao Wu (Nagoya University)
#  MIT License (https://opensource.org/licenses/MIT)

"""Indexing-related functions."""

import torch
from torch.nn import ConstantPad1d as pad1d


def pd_indexing(x, d, dilation, batch_index, ch_index):
    """Pitch-dependent indexing of past and future samples.

    Args:
        x (Tensor): Input feature map (B, C, T).
        d (Tensor): Input pitch-dependent dilated factors (B, 1, T).
        dilation (Int): Dilation size.
        batch_index (Tensor): Batch index
        ch_index (Tensor): Channel index

    Returns:
        Tensor: Past output tensor (B, out_channels, T)
        Tensor: Future output tensor (B, out_channels, T)

    """
    (_, _, batch_length) = d.size()
    dilations = d * dilation

    # get past index
    idxP = torch.arange(-batch_length, 0).float()
    if torch.cuda.is_available():
        idxP = idxP.cuda()
    idxP = torch.add(-dilations, idxP)
    idxP = idxP.round().long()
    maxP = -((torch.min(idxP) + batch_length))
    assert maxP >= 0
    idxP = (batch_index, ch_index, idxP)
    # padding past tensor
    xP = pad1d((maxP, 0), 0)(x)

    # get future index
    idxF = torch.arange(0, batch_length).float()
    if torch.cuda.is_available():
        idxF = idxF.cuda()
    idxF = torch.add(dilations, idxF)
    idxF = idxF.round().long()
    maxF = torch.max(idxF) - (batch_length - 1)
    assert maxF >= 0
    idxF = (batch_index, ch_index, idxF)
    # padding future tensor
    xF = pad1d((0, maxF), 0)(x)

    return xP[idxP], xF[idxF]


def index_initial(n_batch, n_ch, tensor=True):
    """Tensor batch and channel index initialization.

    Args:
        n_batch (Int): Number of batch.
        n_ch (Int): Number of channel.
        tensor (bool): Return tensor or numpy array

    Returns:
        Tensor: Batch index
        Tensor: Channel index

    """
    batch_index = []
    for i in range(n_batch):
        batch_index.append([[i]] * n_ch)
    ch_index = []
    for i in range(n_ch):
        ch_index += [[i]]
    ch_index = [ch_index] * n_batch

    if tensor:
        batch_index = torch.tensor(batch_index)
        ch_index = torch.tensor(ch_index)
        if torch.cuda.is_available():
            batch_index = batch_index.cuda()
            ch_index = ch_index.cuda()
    return batch_index, ch_index
